#
# sip.tcl
# general sip procedures used for incoming SIP requests and responses
# - read SIP and SDP parameters of a SIP message
# - open sessions if invitation was successful
#



#
# extract parameters of a SIP message
# - read all SIP and SDP parameters in msg
# - fill request or response array with extracted parameters
# - return "1" if msg should be ignored, otherwise "0"
#
proc sip_extract_parameters {call_id msg_type msg} {
  global request response conf

  foreach line $msg {
    # check SIP header lines
    set index [string first ":" $line]
    set tag [string range $line 0 $index]
    set value [string trimleft [string range $line [incr index] end]]
    switch [string toupper $tag] {
      V: -
      VIA: {
        lappend ${msg_type}($call_id,via) $value
      }
      F: -
      FROM: {
        if [string match "sip://*" $value] {
          set value [string range $value 6 end]
        }
        set ${msg_type}($call_id,from) $value
      }
      T: -
      TO: {
        if [string match "sip://*" $value] {
          set value [string range $value 6 end]
        }
        set ${msg_type}($call_id,to) $value
      }
      C: -
      CONTENT_TYPE: {
        if {$value != "application/sdp"} {
          # isc can only handle sdp payloads, ignore msg          
          return 1
        }
      }
      L: -
      CONTENT-LENGTH: {
        set ${msg_type}($call_id,length) $value
      }
      E: -
      CONTENT-ENCODING: {
        set ${msg_type}($call_id,encoding) $value
      }
      H: -
      CONTACT-HOST: {
        set ${msg_type}($call_id,host) $value
      }
      M: -
      LOCATION: {
        if [string match "sip://*" $value] {
          set value [string range $value 6 end]
        }
        lappend ${msg_type}($call_id,location) $value
      }
      R: -
      REASON: {
        foreach r [split $value ","] {
          lappend ${msg_type}($call_id,reason) $r
        }
      }
      default {
        # check SDP lines
        if [string match "?=*" $line] {
          lappend ${msg_type}($call_id,payload) $line
        }
        set index [string first "=" $line]
        set tag [string range $line 0 $index]
        set value [string range $line [incr index] end]
        switch $tag {
          a= {
            set ${msg_type}($call_id,attribute) $value
          }
          b= {
          }
          c= {
            set addr [lindex [split [lindex $value 2] /] 0]
            set ttl [lindex [split [lindex $value 2] /] 1]
            # check whether it's connection data for the whole session 
            # or for a media description
            if {[info exists ${msg_type}($call_id,media_list)]} {
              # create new media description for the last media type 
              # specified in the SDP payload
		set old_media_description [lindex [subst $[subst $msg_type]($call_id,media_list)] end]
		set new_media_description [lreplace $old_media_description 2 3 $addr $ttl]
		set ${msg_type}($call_id,media_list) [lreplace [subst $[subst $msg_type]($call_id,media_list)] end end $new_media_description]
            } else {
              # connection data for the whole session
              set ${msg_type}($call_id,session_addr) $addr
              set ${msg_type}($call_id,ttl) $ttl
            }
          }
          e= {
            set ${msg_type}($call_id,email) $value
          }
          i= {
            # check whether it's information data for the whole session 
            # or for a media description
            if {[info exists ${msg_type}($call_id,media_list)]} {
              # create new media description for the last media type 
              # specified in the SDP payload
              set old_media_description [lindex [subst $[subst $msg_type]($call_id,media_list)] end]
	      lappend old_media_description $value 
              set ${msg_type}($call_id,media_list) [lreplace [subst $[subst $msg_type]($call_id,media_list)] end end $old_media_description]
            } else {
              # information data for the whole session
              set ${msg_type}($call_id,information) $value
            }
          }
          k= {
            set ${msg_type}($call_id,key) $value
          }
          m= {
            # if there was no connetion data line ("c=") set session_addr 
            # and ttl to " "
            if ![info exists ${msg_type}($call_id,session_addr)] {
              set ${msg_type}($call_id,session_addr) " "
            }
            if ![info exists ${msg_type}($call_id,ttl)] {
              set ${msg_type}($call_id,ttl) " "
            }
            set media [lindex $value 0]
            set port [lindex $value 1]
            set code [lrange $value 3 end]
            # get names of encodings specified by their RTP/AVP profile
            set code_list ""
            foreach c $code {
              lappend code_list [lindex [get_code $c] 0]
            }
            # create new entry for media_list
	    set ml $media
	    lappend ml $port
	    lappend ml [subst $[subst ${msg_type}]($call_id,session_addr)]
	    lappend ml [subst $[subst ${msg_type}]($call_id,ttl)]
	    lappend ml $code_list
            lappend ${msg_type}($call_id,media_list) $ml
          }
          o= {
            set ${msg_type}($call_id,owner) $value
          }
          p= {
            set ${msg_type}($call_id,phone) $value
          }
          s= {
            # if there are active session, check whether 
            # session_name and conf(name) are equal
            set ${msg_type}($call_id,session_name) $value
            if {[llength $conf(sessions)] && ($value != $conf(name))} {
              # ignore SIP message
              return 1
            }
          }
          t= {
            lappend ${msg_type}($call_id,time) $value
          }
          u= {
            set ${msg_type}($call_id,uri) $value
          }
          v= {
            set ${msg_type}($call_id,version) $value
          }
          z= {
            set ${msg_type}($call_id,zone) $value
          }
        }
      }
    }
  }
  return 0
} ;# sip_extract_parameters



#
# creates requested sessions
# invoked when 200 response received or SIP request was accepted
# 
proc sip_open_sessions {} {
  global request conf Session session response

  set call_id [lindex $request(sessions_to_create) 0]
  catch {after cancel $request($call_id,ringing)}
  wm deiconify .
  catch {destroy .conference_1}
  conference_init_2


  foreach media $request($call_id,media_list) {
      set type  [lindex $media 0]
      set port  [lindex $media 1]
      set addr  [lindex $media 2]
      # unicast invitation to my host ?
    if {[info exists response($call_id,host)]} {
      if {$addr == [idstring %h] || $addr == [idstring %d] || $addr == [idstring %i]} {
        # open session to the contact host of the invitation
        set addr $response($call_id,host)
      }
    }
      set ttl   [lindex $media 3]
      set codes [lindex $media 4]

    set conf(started) 0
    set Session(type)      $type
    set Session(media)     [string toupper [get_code [get_profile [lindex $codes 0]]]] 
    set Session(host)      $addr
    set Session(ttl)       $ttl
    set Session(port)      $port
    set Session(transport) RTP
    set ses "$Session(type)/$Session(host):$Session(port)"
    set session($ses,media) $Session(media)
    set conf(name) $request($call_id,session_name)

    if {[lsearch $conf(sessions) $ses] == -1} {
      session_create
    }
  }
  catch {destroy .response[noperiod $call_id]}
  set request(sessions_to_create) [lreplace $request(sessions_to_create) 0 0]
  if {[llength $request(sessions_to_create)] > 0} {
    sip_open_sessions
  }
} ;# sip_open_sessions



proc sip_sound {file_name} {
  global env conf play


    if {[info exists env(ISC_AUDIO)] && ![catch {set a [open /dev/audio {NONBLOCK RDWR}]}]} {
	close $a
	catch {exec cat $env(ISC_AUDIO)/audio/$file_name > /dev/audio &}
    } else {
	bell; after 300; bell; after 300; bell
    }
} ;# sip_sound


