/*
 *  udp.c --
 *
 *       UDP related functions (open connection, read and write, close)
 *
 *      1996 Dorgham Sisalem
 *  See the copyright file for copyright information
 */
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/fcntl.h>
#include "/usr/include/netinet/in.h"  /* struct sockaddr_in */
#include <arpa/inet.h>   /* inet_ntoa() */
#include <netdb.h>
#include <errno.h>
#include <values.h>
#include <unistd.h>      /* select(), perror() */
#include <string.h>
#include "types.h"
#include "notify.h"
#include "udp.h"


static  struct sockaddr_in serv_addr;
/*
 *  Receive a packet
 */

char* UDP_receive(int sock)
{
    int cli_len, len,i;
    struct sockaddr_in cli_addr;
    char buf[1000];

    for(i=0;i<sizeof(buf);i++)
      buf[i]=0;
    cli_len = sizeof(cli_addr);
    
    len = recvfrom(sock,buf, MAX_PKT_LEN, 0,
		  (struct sockaddr *)&cli_addr, &cli_len);
    return(buf);

}
/*
 *  transmit a packet. Allowes to specify two buffers for scattered write
 *  buf2 must be NULL if only one buffer is necessary
 */
int UDP_write(int sock,char *buf1, int len1, char *buf2, int len2)
{
  struct msghdr msg;
  struct iovec iov[2];

  iov[0].iov_base = buf1;
  iov[0].iov_len = len1;
  iov[1].iov_base = buf2;
  iov[1].iov_len = len2;
  msg.msg_iov = iov;
  msg.msg_iovlen = 2;
  msg.msg_name = (caddr_t *) &serv_addr;
  msg.msg_namelen = sizeof(serv_addr);
#if defined(linux)
  msg.msg_control = NULL;
  msg.msg_controllen = 0;
  msg.msg_flags = 0;
#else
  msg.msg_accrights = NULL;
  msg.msg_accrightslen = 0;
#endif
  sendmsg(sock, &msg, 0);
  return len1 + len2;
}


/* 
 * open unicast or multicast UDP connection
 *
 *
*/
int UDP_connect(char *host, int port, u_int8 ttl)
{ 
  int reuse = 1;  /* reuse address */
  struct sockaddr_in sin;
  int multicast;
  struct in_addr sin_addr;
  int sock;
  
  sin_addr.s_addr = inet_addr(host);

  serv_addr.sin_port = htons(port);
  serv_addr.sin_family = AF_INET;
  serv_addr.sin_addr.s_addr = htonl(sin_addr.s_addr);

  multicast = IN_MULTICAST(htonl(serv_addr.sin_addr.s_addr));
  memset(&sin, 0, sizeof(sin));		      
  sin.sin_family = AF_INET;
  sin.sin_addr.s_addr = multicast ? htonl(sin_addr.s_addr) : htonl(INADDR_ANY);

  sin.sin_port = htons(port);
		      
  /* Open a UDP socket */
   
  sock = socket(PF_INET, SOCK_DGRAM, 0);
  if (sock < 0) {
    perror("can not open socket");
    return(-1);
  }

  if (setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, (char *)&reuse,
		 sizeof(reuse)) == -1){
  	 perror("setsockopt: reuseaddr");
	return(-1);
  }

 if (multicast) {
    struct ip_mreq mreq;
    int one = 1;
    u_int8 loop =1;
    int size, len;
    mreq.imr_multiaddr.s_addr = htonl(sin_addr.s_addr);
    mreq.imr_interface.s_addr = htonl(INADDR_ANY);

    if (setsockopt(sock, IPPROTO_IP, IP_ADD_MEMBERSHIP,
		   (char *)&mreq, sizeof(mreq)) < 0) {
      perror("IP_ADD_MEMBERSHIP");
      return(-1);
    }

    if (setsockopt(sock, IPPROTO_IP, IP_MULTICAST_TTL, 
		   (char *)&ttl,sizeof(u_int8)) < 0) {
      perror("IP_MULTICAST_TTL");
      return(-1);
    }
    
    if (setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, (char *)&one,
		   sizeof(one)) == -1) {
      perror("setsockopt: reuseaddr");
      return(-1);
    }
  
  
    if (setsockopt(sock,IPPROTO_IP,
		   IP_MULTICAST_LOOP, (char *)&loop, sizeof(u_int8)) < 0) {
      perror("IP_MULTICAST_LOOP");
      return(-1);
    }
    /*
     * Make sure the send/recv buffers are of a reasonable size
     * (this is a problem under Solaris 2.x) -- the default size is 0.
     */
    size = 65536;
    len = sizeof(int);
    getsockopt(sock, SOL_SOCKET, SO_SNDBUF, (char *)&size, &len);
    if (size < 65536) {
      size = 65536;
    }
    len = sizeof(int);
    setsockopt(sock, SOL_SOCKET, SO_SNDBUF, (char *)&size, 
	       sizeof(int));
    getsockopt(sock, SOL_SOCKET, SO_RCVBUF, (char *)&size, &len);
    if (size < 65536) {
      size = 65536;
    }
    setsockopt(sock,SOL_SOCKET,SO_RCVBUF,(char *)&size, sizeof(int));
    
    
  } /* if multicast */


  if (bind(sock, (struct sockaddr *)&sin, sizeof(sin)) < 0) {
      /*
       * Binding to multicast address fails if VJ's multicast mods have
       * not been installed, so we try again with INADDR_ANY.
       */
      if (errno == EADDRNOTAVAIL) {
        sin.sin_addr.s_addr = INADDR_ANY;
	if (bind(sock, (struct sockaddr *)&sin, sizeof(sin)) < 0){
	  perror("bind");
	  return(-1);
	}
      }
    } /* if bind */

  return (sock);
}





/*
 *  close sockets and remove file event handler
 */ 
int  UDP_close(int sock)
{
  if(sock>0)
    close(sock);
  return 1;
}

