/*
hp_audio  --  audio interface routines for HP 700 (HP-UX)

Audio I (9000/710):
8000/1/PCM only.
Audio ID:              1
Max bits per sample:  16
Number of channels:    1
Receive gain:        -60 ... 14
Transmit gain:       -60 ... 14
Monitor gain:        -54 ...  0

Does NOT support RESET_OVRANGE option.

Audio II (9000/725):
8000 11025 16000 22050 32000 44100 48000
for both PCM and L16, 1 and 2 channels.
Audio ID:              2
Max bits per sample:  16
Number of channels:    2
Receive gain:          0 ... 22 
Transmit gain:       -84 ...  9
Monitor gain:        -84 .... 0

All combinations are supported, but receive and transmit must
have the same values. Apparently, once audio data has been
read from the device, the values cannot be changed.

The HP audio system (apparently) does not have the ability to
send a signal to all applications having an audio device open when
changes occur (unlike SunOS). We simulate the behavior by catching
SIGPOLL (SIGIO).

Copyright 1994 by Henning Schulzrinne; all rights reserved
*/



#include "types.h"
#include <sys/audio.h>   /* HP-specific */
#include <fcntl.h>       /* O_RDONLY */
#include <unistd.h>      /* close(), ioctl() */
#include <errno.h>       /* EIO */
#include <signal.h>      /* SIGPOLL */
#include <memory.h>      /* memcmp() */
#include "audio_generic.h"
#include "ahw.h"
#include "ansi.h"

#define AUDIO_DATA "/dev/audio"
#define AUDIO_CTL  "/dev/audioCtl"

static int ac = -1;                 /* audio control device */
static int au = -1;                 /* audio data descriptor */ 
static audio_state_t as;
static signal_func_t track;         /* handler that tracks audio state */
static struct audio_describe AD;    /* HP hardware capabilities */

static struct audio_select_thresholds thresholds = {160, 160};

int ahw_changed = 1;

/*
* Set up audio format for open audio device based on audio descriptor.
*/
static int ahw_setup(audio_descr_t *ad)
{
  int v;

  if (au >= 0) {
    if      (ad->encoding == AE_PCMU) v = AUDIO_FORMAT_ULAW;
    else if (ad->encoding == AE_PCMA) v = AUDIO_FORMAT_ALAW;
    else if (ad->encoding == AE_L16)  v = AUDIO_FORMAT_LINEAR16BIT;
    else return -1;
    if (ioctl(au, AUDIO_SET_DATA_FORMAT, v) < 0) {
      perror("SET_DATA_FORMAT");
      return -1;
    }
    if (ioctl(au, AUDIO_SET_SAMPLE_RATE, ad->sample_rate) < 0) {
      perror("SET_SAMPLE_RATE");
      return -1;
    }
    if (ioctl(au, AUDIO_SET_CHANNELS, ad->channels) < 0) {
      perror("SET_CHANNELS");
      return -1;
    }
  }
  return 0;
} /* ahw_setup */

/*
* Open audio control device. Return -1 if error.
*/
static int ahw_open_control(void)
{
  if (ac < 0) {
    if ((ac = open(AUDIO_CTL, O_RDWR, 0)) == -1) return -1;
    /* capabilities */
    if (ioctl(ac, AUDIO_DESCRIBE, &AD) < 0) return -1;
  }
  return ac;
} /* ahw_open_control */

/*
* Open audio hardware and set file descriptor to non-blocking.  Return
* file descriptor number or -1 on error.  If the output device is
* busy, we don't fail, but would try again later, thus grabbing the
* audio device as quickly as possible.  This allows short
* announcements (e.g., biff) to grab the audio device without crashing
* this program.  If the other program changes the device
* characteristics, there is potential for disaster.  Several nevot's
* can run concurrently if the nevot.talk is used to disable the
* recording device (iff there is no audio output waiting).
*/
int ahw_open(int ndelay)
{
  int fd_flags;  /* fcntl flags */

  /* open audio data stream */
  if (au < 0) {
    if ((au = open(AUDIO_DATA, O_RDWR | O_NDELAY)) < 0) {
      return -1;
    }
    if (ndelay) {
      fd_flags = fcntl(au, F_GETFL, 0);
      (void)fcntl(au, F_SETFL, fd_flags|O_NDELAY);
    }

    /* set audio format */
    ahw_setup(&as.play.ad);

    /* set read/write buffer sizes */
    if (ioctl(au, AUDIO_SET_RXBUFSIZE, 16384) == -1) return -1;
    if (ioctl(au, AUDIO_SET_TXBUFSIZE, 16384) == -1) return -1;
    if (ioctl(au, AUDIO_SET_SEL_THRESHOLD, (int *)&thresholds) == -1) return -1;
  }

  return au;
} /* ahw_open */

/*
* Reset audio hardware, e.g., after error.
* Since OVRANGE is defined for Audio II only (whatever that may be),
* we try it first and then retry without it on failure.
*/
static int ahw_reset(void)
{
  int reset = RESET_TX_BUF|RESET_TX_UNF|RESET_RX_BUF|RESET_RX_OVF;

  printf("reset\n");
  if (ioctl(au, AUDIO_RESET, reset | RESET_OVRANGE) < 0) {
    return ioctl(au, AUDIO_RESET, reset);
  }
  return 0;
} /* ahw_reset */


/*
* Signal handler.  Call actual handler and then reestablish signal
* handler.
*/
static void ahw_signal(int sig)
{
  (*track)(sig);
  signal(SIGPOLL, ahw_signal);
} /* ahw_signal */


/*
* Close audio hardware.  Return old (now closed) input file descriptor
* to allow disabling input handlers.
*/
int ahw_close(void)
{
  int au_old = au;

  if (au >= 0) {
    close(au);
    au = -1;
  }
  if (ac >= 0) {
    close(ac);
    ac = -1;
  }
  return au_old;
} /* ahw_close */

/*
* Set audio hardware. Audio device does not have to be open.
* Return 0 if ok, -1 on error.
* pr     in   0: read, 1: write
* p      in   parameter to set
* value  in   value to be used; -1: don't change
*/
int ahw_set(int pr, audio_set_t p, int value)
{
  int success = 0;       /* return status */
  int v = 0;
  struct audio_gain ag;  /* current audio gain */

  ahw_open_control();

  switch(p) {
  case AS_port:
    if (pr == 0) {
      if      (value == AP_in_line)       v = AUDIO_IN_LINE;
      else if (value == AP_in_microphone) v = AUDIO_IN_MIKE;
      else return -1;
      success = ioctl(ac, AUDIO_SET_INPUT, v);
    }
    else {
      if (value & AP_out_speaker)   v |= AUDIO_OUT_SPEAKER;
      if (value & AP_out_headphone) v |= AUDIO_OUT_HEADPHONE;
      if (value & AP_out_line)      v |= AUDIO_OUT_LINE;
      if (value & ~(AP_out_speaker|AP_out_headphone|AP_out_line))
        return -1;
      success = ioctl(ac, AUDIO_SET_OUTPUT, v);
    }
    break;

  case AS_gain_left:
  case AS_gain_right:
  case AS_gain:
    /*
     * Set ag to current value. This tracks any external changes
     * caused by other applications modifying audio device
     * characteristics.
     */ 
    if (ioctl(ac, AUDIO_GET_GAINS, &ag) < 0) return -1;

    switch(pr) {
    case 0:  /* record */
      v = value/255. * (AD.max_receive_gain - AD.min_receive_gain) +
        AD.min_receive_gain;
      if (p == AS_gain)
        ag.cgain[1].receive_gain = ag.cgain[0].receive_gain = v;
      else if (p == AS_gain_left)
        ag.cgain[0].receive_gain = v;
      else if (p == AS_gain_right)
        ag.cgain[1].receive_gain = v;
      break;
    case 1:  /* play */
      v = value/255. * (AD.max_transmit_gain - AD.min_transmit_gain) +
        AD.min_transmit_gain;
      if (p == AS_gain)
        ag.cgain[1].transmit_gain = ag.cgain[0].transmit_gain = v;
      else if (p == AS_gain_left)
        ag.cgain[0].transmit_gain = v;
      else if (p == AS_gain_right)
        ag.cgain[1].transmit_gain = v;
      break;
    case 2: /* monitor */
      v = value/255. * (AD.max_monitor_gain - AD.min_monitor_gain) +
        AD.min_monitor_gain;
      if (p == AS_gain)
        ag.cgain[1].monitor_gain = ag.cgain[0].monitor_gain = v;
      else if (p == AS_gain_left)
        ag.cgain[0].monitor_gain = v;
      else if (p == AS_gain_right)
        ag.cgain[1].monitor_gain = v;
      break;
    default:
      break;
    }
    if (p == AS_gain)
      ag.channel_mask = AUDIO_CHANNEL_LEFT | AUDIO_CHANNEL_RIGHT;
    else if (p == AS_gain_left)
      ag.channel_mask = AUDIO_CHANNEL_LEFT;
    else if (p == AS_gain_right)
      ag.channel_mask = AUDIO_CHANNEL_RIGHT;
    success = ioctl(ac, AUDIO_SET_GAINS, &ag);
    break;

  case AS_balance:
    break;

  case AS_gain_step:
    break;

  case AS_fillpoint:
    if (pr)
      thresholds.read_threshold = value;
    else
      thresholds.write_threshold = value;
    if (au >= 0) {
      success = ioctl(au, AUDIO_SET_SEL_THRESHOLD, (int *)&thresholds);
      break;
    }
  }
  return success;
} /* ahw_set */

/*
* Establish function that tracks audio device changes.
*/
void ahw_track(signal_func_t func)
{
  if (track == func) return;
  track = func;
  if (track) {
    signal(SIGPOLL, ahw_signal);
  }
  else {
    signal(SIGPOLL, SIG_IGN);
  }
} /* ahw_track */


/*
* Set audio device to best approximation of 'record' and 'play'. 
* Since playing and recording and the individual parameters are not
* independent, we have to set them together.  On return, 'record' and
* 'play' contain the setting used.  Record overrides play.  For HPs,
* they will be forced to be the same.  Return -1 on error (illegal
* sampling rate for device).
*/
int ahw_fmt(audio_descr_t *record, audio_descr_t *play)
{
  int i;

  /* need AD */
  ahw_open_control();

  /*
   * Non-L16 encodings are all mapped to PCMU.
   */
  if (AD.max_bits_per_sample == 16) {
    if (record->encoding != AE_L16) {
      record->encoding = AE_PCMU;
    }
  }
  else record->encoding = AE_PCMU;

  /*
   * For a mono device, we force mono.
   */
  if (AD.nchannels == 1) record->channels = 1;

  for (i = 0; i < AD.nrates; i++) {
    if (record->sample_rate == AD.sample_rate[i]) goto found;
  }
  return -1;  /* no matching sample rate */

found:
  *play = *record;

  /*
   * Check if any change. If not, return. Avoids unnecessary ioctl()
   * and errors when changing running audio device.
   */
  if (memcmp(record, &as.record.ad, sizeof(audio_descr_t)) == 0 &&
      memcmp(play,   &as.play.ad,   sizeof(audio_descr_t)) == 0)
    return 0;

  as.play.ad   = *play;
  as.record.ad = *record;

  return ahw_setup(&as.play.ad);
} /* ahw_fmt */

/*
* Return number of samples in play queue.
* If the number is negative, an underflow has occurred.
* Return -1 if error.
*/
int ahw_write_queue(int *underflow)
{
  struct audio_status status;

  /* current device status (errors) */
  if (ioctl(ac, AUDIO_GET_STATUS, &status) < 0) {
    *underflow = 1;
    return -1;
  }
  else {
    *underflow = 0;
    return status.transmit_buffer_count;
  }
} /* ahw_write_queue */


/*
* Return occupancy of record queue in samples.
* Return negative value if error.
*/
int ahw_read_queue(int *overflow)
{
  struct audio_status status;

  /* current device status (errors) */
  if (ioctl(ac, AUDIO_GET_STATUS, &status) < 0) {
    *overflow = 1;
    return -1;
  }
  else {
    *overflow = 0;
    return status.receive_buffer_count;
  }
} /* ahw_read_queue */

/*
* Get current hardware audio status.
* Return -1 if error, 0 if ok.
*/
int ahw_get(audio_state_t *as)
{
  int v;
  int i;
  struct audio_gain gains;
  struct audio_select_thresholds thresh;
  struct audio_status status;

  ahw_open_control();

  /* current device status (errors) */
  if (ioctl(ac, AUDIO_GET_STATUS, &status) < 0) return -1;
  as->play.errors   = status.transmit_underflow_count;  
  as->record.errors = status.receive_overflow_count;

  /* write port */
  as->play.port = 0;
  if (ioctl(ac, AUDIO_GET_OUTPUT, &v) < 0) return -1;
  if (v & AUDIO_OUT_SPEAKER)   as->play.port |= AP_out_speaker;
  if (v & AUDIO_OUT_HEADPHONE) as->play.port |= AP_out_headphone;
  if (v & AUDIO_OUT_LINE)      as->play.port |= AP_out_line;

  /* read port */
  as->record.port = 0;
  if (ioctl(ac, AUDIO_GET_INPUT, &v) < 0) return -1;
  if (v & AUDIO_IN_MIKE) as->record.port |= AP_in_microphone;
  if (v & AUDIO_IN_LINE) as->record.port |= AP_in_line;

  /* input, output and monitor gain */
  if (ioctl(ac, AUDIO_GET_GAINS, &gains) < 0) return -1;
  for (i = 0; i < 2; i++) {
    as->play.gain[i]    = 
      (gains.cgain[i].transmit_gain - AD.min_transmit_gain) /
      (double)(AD.max_transmit_gain - AD.min_transmit_gain) * 255;
    as->record.gain[i]  = 
      (gains.cgain[i].receive_gain - AD.min_receive_gain) /
      (double)(AD.max_receive_gain - AD.min_receive_gain) * 255;
    as->monitor_gain[i] =
      (gains.cgain[i].monitor_gain - AD.min_monitor_gain) /
      (double)(AD.max_monitor_gain - AD.min_monitor_gain) * 255;
  }

  /* buffer size */
  if (ioctl(ac, AUDIO_GET_RXBUFSIZE, &as->record.buffer_size) < 0) return -1;
  if (ioctl(ac, AUDIO_GET_TXBUFSIZE, &as->play.buffer_size) < 0) return -1;

  /* fillpoint */
  if (ioctl(ac, AUDIO_GET_SEL_THRESHOLD, &thresh) < 0) return -1;
  as->play.fillpoint     = thresh.write_threshold;
  as->record.fillpoint   = thresh.read_threshold;

  for (i = 0; i < 2; i++) {
    audio_ws_t *ws = (i == 0) ? &as->record : &as->play;

    ws->active       = 0;
    ws->waiting      = 0;
  }

  /* encoding */
  if (ioctl(ac, AUDIO_GET_DATA_FORMAT, &v) < 0) return -1;
  if (v == AUDIO_FORMAT_ULAW) {
    as->play.ad.encoding   = AE_PCMU;
  }
  else if (v == AUDIO_FORMAT_ALAW) {
    as->play.ad.encoding   = AE_PCMA;
  }
  else if (v == AUDIO_FORMAT_LINEAR16BIT) {
    as->play.ad.encoding   = AE_L16;
  }
  else {
    as->play.ad.encoding   = AE_MAX;
  }

  /* sample rate */
  if (ioctl(ac, AUDIO_GET_SAMPLE_RATE, &v) < 0) return -1;
  as->play.ad.sample_rate   = v;

  /* channels */
  if (ioctl(ac, AUDIO_GET_CHANNELS, &v) < 0) return -1;
  as->play.ad.channels   = v;

  /* Play and record must be the same. */
  as->record.ad = as->play.ad;

  return 0;
} /* ahw_get */

/*
* Return > 0 if another process is waiting for audio input or output.
*/
int ahw_waiting(void)
{
  return 0;
} /* ahw_waiting */


/*
* Write audio buffer to audio device. Return number of bytes written
* or -1 on error.
*/
int ahw_write(char *buffer, int bytes)
{
  int left = bytes;
  int nwritten;

  if (bytes == 0) return bytes;
  if (au >= 0) {
    while (left > 0) {
      nwritten = write(au, buffer, left);
      left   -= nwritten;
      buffer += nwritten; 
    }
  }
  return bytes;
} /* ahw_write */

/*
* Read audio samples from audio device. Return number of bytes read,
* -1 on error. Call recursively if audio device had to be reset, but
* limit recursion depth to 10.
* buffer  out   buffer to receive audio data
* bytes   in    number of audio bytes to read
*/
int ahw_read(char *buffer, int bytes)
{
  int nread;
  int left = bytes;
  static int count;  /* limit recursion */

  while (left > 0 && count < 10) {
    if ((nread = read(au, buffer, left)) == -1) {
      if (errno == EIO) {
        ahw_reset();
        count++;
        return ahw_read(buffer, left);
      }
      bytes = -1;
      break;
    }
    else {
      left   -= nread;
      buffer += nread;
    }
  }
  count = 0;
  return bytes;
} /* ahw_read */
