/* 3.28
udp  --  UDP and multicast-UDP related functions.

UDP has one socket for outgoing data and two for incoming, one for data,
one for control.

If sockets are connected (which they are currently not), we get
ECONNREFUSED if there's nobody listening at that port.  We ignore that
error.

Copyright 1994 by Henning Schulzrinne; all rights reserved
*/

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <errno.h>
#include <unistd.h>     /* close() */
#include <stdio.h>
#include <stdlib.h>     /* malloc() */
#include <sys/uio.h>    /* sendmsg() */
#include "tsap.h"
#include "mbuf.h"
#include "msghdr.h"
#include "notify.h"
#include "udp.h"
#include <ansi.h>



/*
* Create UDP descriptor *u for input and output.  For multicast, use
* address 'tsap', otherwise ignore address.  for the local address use tsapl
* Return error indication.
*/
int UDP_connect(tsap_t *tsap, tsap_t *tsapl, Notify_func_input func, 
		Notify_client client, void **u_pt)
{
  int one = 1;
  udp_t *u;
  struct sockaddr_in sin;
  struct in_addr interface;
  int v; /* bind result */

  /* no transport data structure */
  u = (udp_t *)calloc(1, sizeof(udp_t));
  u->in        = -1;
  u->out       = -1;
  u->multicast = tsap_ismulticast(tsap);
  u->addr      = tsap_ip4(tsap);
  u->port      = tsap_getport(tsap);

  
  interface.s_addr=tsap_ip4(tsapl);  /* local address */

  /* create input socket if it don't exist already */
  if (u->in < 0) {
    u->in = socket(PF_INET, SOCK_DGRAM, 0);
    if (u->in < 0) return -1;
    /* bind it for control and data */
    sin.sin_family      = AF_INET;
    sin.sin_port        = tsap_getport(tsap);
    if(u->multicast>0)
      sin.sin_addr.s_addr = tsap_ip4(tsap);
    else
      sin.sin_addr.s_addr = tsap_ip4(tsapl); /* bind address to local interface
					      */
    /* sin.sin_addr.s_addr = u->multicast ? tsap_ip4(tsap) : INADDR_ANY;*/
    /*
     * SO_REUSEADDR allows two sessions to bind to same port.
     * For unicast, only the last to bind will receive packets, but
     * earlier ones will pick up after the last arrival leaves.
     * For multicast, this should be irrelevant (both get all packets).
     * This is also the vat behavior.
     */
    if (setsockopt(u->in, SOL_SOCKET, SO_REUSEADDR, 
        (char *)&one, sizeof(one)) < 0) goto error;

    /* handle the multihomes hosts case */
    if (u->multicast) {
#ifdef IP_ADD_MEMBERSHIP
      struct ip_mreq mreq;      /* multicast group */

      mreq.imr_multiaddr.s_addr = u->addr;
      mreq.imr_interface.s_addr = interface.s_addr;
      if (setsockopt(u->in, IPPROTO_IP, IP_MULTICAST_IF,
		   (char *)&interface, sizeof(interface)) < 0) {
	perror("IP_MULTICAST_IF");
	return(NULL);
      }
      /*mreq.imr_interface.s_addr = htonl(INADDR_ANY);*/
      if (setsockopt(u->in, IPPROTO_IP,
        IP_ADD_MEMBERSHIP, (char *)&mreq, sizeof(mreq)) < 0) {
        goto error;
      }
#else
      errno = ENOPROTOOPT;
      goto error;
#endif
    }
    if ((v = bind(u->in, (struct sockaddr *)&sin, sizeof(sin))) < 0) {
      /*
       * Binding to multicast address fails if VJ's multicast mods have
       * not been installed, so we try again with INADDR_ANY.
       */
      if (errno == EADDRNOTAVAIL) {
        sin.sin_addr.s_addr = INADDR_ANY;
        v = bind(u->in, (struct sockaddr *)&sin, sizeof(sin));
      }
      if (v < 0) goto error;
    } /* if bind */
    notify_set_input_func((Notify_client)client, func, u->in);
  } /* if */

  /* create output socket */
  if (u->out < 0) {
    u->out = socket(PF_INET, SOCK_DGRAM, 0);
    if (u->out < 0) return -1; 

    if (u->multicast) {
      struct ip_mreq mreq;      /* multicast group */
#ifdef IP_MULTICAST_LOOP	
      u_int8 loop = 0;          /* output loop back */

      if (setsockopt(u->out, IPPROTO_IP,
                    IP_MULTICAST_LOOP, (char *)&loop, sizeof(loop)) < 0) {
        goto error;
      }
#endif
      mreq.imr_multiaddr.s_addr = u->addr;
      mreq.imr_interface.s_addr = interface.s_addr;
      if (setsockopt(u->out, IPPROTO_IP, IP_MULTICAST_IF,
		     (char *)&interface, sizeof(interface)) < 0) {
	perror("IP_MULTICAST_IF");
	return(NULL);
      }
    } /* multicast */
  } /* if */
  *u_pt = u;
  return 0;

error:
  if (u->in >= 0) {
    close(u->in);
    u->in = -1;
  }
  if (u->out >= 0) {
    close(u->out);
    u->out = -1;
  }
  *u_pt = (void *)0;
  free(u);
  return -1;
} /* UDP_connect */


/*
* Read single datagram from socket 'sock' into 'buf'.
* Return number of bytes read.
*/
int UDP_read(int sock, void *buf, int size, tsap_t *tsap)
{
  int rcount;

  rcount = recvfrom(sock, buf, size, 0, TSAP_SAP(tsap), (int *)&tsap->addr.len);
  return rcount;
} /* UDP_read */


/*
* Write single datagram to 'tsap', gathering mbufs into I/O vector. If 'u' is
* zero, return immediately.
*/
int UDP_write(void *u_pt, mbuf_t *mh, tsap_t *tsap)
{
  udp_t *u = (udp_t *)u_pt;
  int sent;
  struct iovec io[10];
  mbuf_t *mb;
  struct msghdr msg;
  int i;

  if (!u) return 0;
  for (mb = mh, i = 0; mb; mb = mb->m_next, i++) {
    io[i].iov_base = mtod(mb, char *);
    io[i].iov_len  = mb->m_len;
  }

  msg.msg_iov = io;
  msg.msg_iovlen = i;
  msg.msg_name = (caddr_t)TSAP_SAP(tsap);
  msg.msg_namelen = sizeof(struct sockaddr);
  msg.msg_accrights = 0;
  msg.msg_accrightslen = 0;

  sent = sendmsg(u->out, &msg, 0);
  return sent;
} /* UDP_write */

/*
* Close UDP "connection" for member 'm' of session 's';
* leave multicast group if 'm' is multicast.
* Called for each session member.
*/
int UDP_close(void *u_pt)
{
  udp_t *u = (udp_t *)u_pt;

  if (!u) return 0;
  if (u->in >= 0) {
    if (u->multicast) {
#ifdef IP_MULTICAST_TTL	
      struct ip_mreq mreq;      /* multicast group */

      mreq.imr_multiaddr.s_addr = u->addr;
      mreq.imr_interface.s_addr = INADDR_ANY;

      if (setsockopt(u->in, IPPROTO_IP, IP_DROP_MEMBERSHIP,
        (char *)&mreq, sizeof(mreq)) < 0) {
        return -1;
      }
#endif
    }
    notify_set_input_func(0, NOTIFY_FUNC_INPUT_NULL, u->in);
    close(u->in);
    u->in = -1;
  }
  if (u->out >= 0) close(u->out);
  free((char *)u);
  return TRUE;
} /* UDP_close */

/*
* Change TTL value of output socket (multicast only).
* Return -1 if error.
*/
int UDP_ttl(void *u_pt, int ttl)
{
  char ttl_char;
  udp_t *u = (udp_t *)u_pt;

  ttl_char = ttl;
  if (u->out >= 0) {
#ifdef IP_MULTICAST_TTL	
    return setsockopt(u->out, IPPROTO_IP, IP_MULTICAST_TTL,
      &ttl_char, sizeof(u_int8));
#endif
  }
  return -1;
} /* UDP_ttl */
