/*
audio  --  system-independent interface functions for audio device.

Definitions:
------------

sample: one audio sample (one per channel); may be smaller or larger
than one byte

sample_rate: number of sample per second (per channel)

channels: "number of interleaved sample units.  For any given time
quantum, the set consisting of 'channels' sample units is called a
sample frame.  Seeks in the data should be aligned to the start of the
nearest sample frame."

Examples:
---------
stereo 8-bit PCM:
0        1        2
|  left  | right  |
| sample | sample |
|  unit  |  unit  |
|      frame      |

stereo 4-bit ADPCM:
0               1               2  byte
| left channel  | right channel |
| sample sample | sample sample | 
|     unit      |     unit      |
|             frame             |

At this point, it seems reasonable to assume that all parties use the
same frame rate; on-the-fly frame rate conversions seem to be difficult
at best. If not, the receiver should use the highest frame rate
employed and interpolate as necessary.

Similarly, the receiver should use the maximum number of channels
requested. This may be difficult to do dynamically since some systems
(SGI, for example) cannot change the number of channels once the
audio device has been opened. 

For playing audio files, off-line conversion to the desired frame
rate, precision and channel count is probably best.

For voice-quality systems, the following scheme is used:
typedef u_int8 buffer_t;
net -> [ADPCM, LPC, ...] -> buffer (mu-law) -> audio 
 
For systems with advanced audio capabilities (multiple frame rates 
and widths):
typedef int16 buffer_t;
net -> [ADPCM, LPC, ...] -> [mu-law to linear] -> buffer (16-bit linear) -> audio

Note that buffering as 16-bit linear samples eliminates linearization
during mixing.

Copyright 1993 by AT&T Bell Laboratories; all rights reserved
*/

#include <stdio.h>
#include <sys/file.h>

#include <fcntl.h>
#include <errno.h>
#include <unistd.h>
#include <stdlib.h>   /* atoi() */

#include "types.h"    /* boolean */
#include "nevot.h"
#include "audio_generic.h"
#include "ahw.h"
#include "list.h"     /* mask_to_list() */
#include "util.h"     /* itoa() */
#include "notify.h"
#include "debug.h"    /* debug() */
#include <ansi.h>     /* architecture-specific include fixes */



short bpu[] = /* bytes per unit */
  {1, 1, 1, 3, 0, 0, 1, 2, 4, 4, 4, 4};
short spu[] = /* samples per unit */
  {1, 2, 2, 8, 0, 0, 1, 1, 1, 1, 1, 1};
int codec_sample_rate[] = /* default codec sample rate */
  {8000, 8000, 8000, 8000, 8000, 8000, 8000, 8000, 8000};
extern codec_t codec[];    /* codec description */

#define OFFSET_1900_1970 (unsigned)2208988800
static int opened;

/*
* Change hardware audio parameters to match incoming/outgoing audio
* data, described by 'ad', for audio input (pr=0) or audio output
* (pr=1).  Modifies 'ad' to fill in missing fields.  Also set
* nevot.rx.ad or nevot.tx.ad to correspond to current actual hardware
* format.
*
* For playback, we need to find a method to avoid constant toggling
* if audio formats differ between sources, in particular since it
* causes the content of the audio buffer to be misinterpreted.
* Could use member with highest quality, but then need a way to
* back off if that member were to leave or become inactive.
*
* Return Audio_OK if no error, Audio_EENC if illegal encoding.
*/
Audio_errno_t audio_change(int pr, audio_descr_t *ad)
{
  audio_descr_t *ac;   /* current encoding */

  if (ad->encoding == AE_MAX) return Audio_EENC;
  ac = pr ? &nevot.rx.ad : &nevot.tx.ad;
  if (memcmp(ac, ad, sizeof(audio_descr_t))) {
    *ac = *ad;
    /*
     * If output format is undefined, set it to input format.
     */
    if (!nevot.rx.ad.channels) nevot.rx.ad = nevot.tx.ad;

    if (ahw_fmt(&nevot.tx.ad, &nevot.rx.ad) < 0) return Audio_ERROR;
    /*
     * Adjust amount of data to be read according to encoding.
     * Only changes if sampling rate changes.
     */
    audio_chunk();
  }

  return Audio_OK;
} /* audio_change */

/*
* Change outgoing audio encoding based on incoming packet.  For
* encoding, change physical play encoding to what seems most sensible
* (anything but L16 to PCMU).  Should change channel count only if
* going from 1 to 2. Returns without doing anything if encoding is invalid.
*/
void audio_out(Audio_member_t m, audio_descr_t *ad)
{
  Audio_errno_t result;

  /* adjust decoder state for member */
  if (m->decoder_state) free(m->decoder_state);
  if (codec[ad->encoding].init)
    m->decoder_state = codec[ad->encoding].init(0, nevot.audio.tick);

  /* Change local audio output. */
  m->audio = *ad;
  if ((result = audio_change(1, ad)) != Audio_OK) {
    (*nevot.event)(m->s, m, Audio_error, result);
  }
} /* audio_out */

/*
* Open audio device.  Return 0 if already open, 1 if newly
* opened, -1 if error.
*/
Audio_errno_t audio_open(void)
{
  int au;
  extern int errno;

  if (opened) return 0;
  au = ahw_open(TRUE);
  if (au < 0) {
    switch (errno) {
    case EBUSY:
      return Audio_EAUDIO_BUSY;
    case EACCES:
      return Audio_EAUDIO_ACCES;
    case EINTR:
      return Audio_EAUDIO_INTR;
    case EIO:
      return Audio_EAUDIO_IO;
    case ENXIO:
      return Audio_EAUDIO_NXIO;
    default:
      return Audio_EAUDIO_OPEN;
    }
  }
  /* DEC AudioFile */
  else if (au == 0) {
    notify_set_periodic_func((Notify_client)1, (Notify_func)tx, 20);
  }
  else {
    notify_set_input_func((Notify_client)0, tx, au);
  }
  tx_start();
  opened = TRUE;
  debug(DEBUG_AUDIO, "audio opened");
  return 1;
} /* audio_open */

/*
* Close audio device (input and output).
* Reset VU meters.
*/
void audio_close(void)
{
  int au;

  if (opened) {
    au = ahw_close();
    /* systems that read audio from file */
    if (au) {
      notify_set_input_func((Notify_client)0, NOTIFY_FUNC_INPUT_NULL, au);
    }
    /* systems that read audio data periodically */
    else {
      notify_set_periodic_func((Notify_client)1, (Notify_func)tx, 0);
    }
    audio_vu(0, 0.);
    audio_vu(1, 0.);
    opened = FALSE;
    debug(DEBUG_AUDIO, "audio closed");
  }
} /* audio_close */

/*
* Close audio device if nobody else is using it either for playing
* audio in ring buffer or for listening to input device.
*/
void audio_close_idle(void)
{
  Audio_session_t s;

  if (nevot.audio.talk == 0) {
    if (nevot.audio_priority <= 1 || 
        (nevot.audio_priority == 2 && nevot.audio.waiting)) {
      ALL_SESSIONS(s) {
        if (ring_left(s->ring) > 0) goto active;
      }
      audio_close();
      /* mark all local/outside members as inactive */
      ALL_SESSIONS(s) talking_clear(s);
    }
  }
  active: ;
} /* audio_close_idle */

/*
* Map audio descriptor to header content number, using mmap 'ad_map'
* (which will depend on the RT protocol in use.)
* If no matching encoding found, return -1.  
*/
int audio_to_header(audio_descr_t *ad, audio_descr_t *ad_map)
{
  int i;

  for (i = 0; i < 64; i++) {
    if (ad->encoding    == ad_map[i].encoding &&
        ad->sample_rate == ad_map[i].sample_rate &&
        ad->channels    == ad_map[i].channels) {
      return i;
    }
  } 
  return -1;
} /* audio_to_header */


/*
* Function invoked when audio device characteristics change.
*/
Notify_value audio_poll(Notify_client client, int sig, Notify_signal_mode when)
{
  Audio_session_t s;
  int active = 0;  /* one or more sessions active? */

  if (ahw_waiting()) {
    nevot.audio.waiting = 1;
    debug(DEBUG_AUDIO, "other process waiting");
    ALL_SESSIONS(s) {
      if (ring_left(s->ring) > 0) {
        active++;
        break;
      }
    }
    if ((nevot.audio_priority == 0) ||
        (!active && (nevot.audio_priority == 2))) {
      audio_close();
    }
  }
  return NOTIFY_DONE;
} /* audio_poll */


/*
* VU meter handlers.
*/
void audio_vu(int i, float value)
{
  (*nevot.event)(0, 0, Audio_vu, i, value);
} /* audio_vu */
void audio_vu_max(int i, float value)
{
  (*nevot.event)(0, 0, Audio_vu_max, i, value);
} /* audio_vu */


/*
* Adjust parameters that depend on the local audio format.  Convert
* chunk size to samples and bytes.  Must be called when
* nevot.?.ad.sample_rate changes
*/
void audio_chunk(void)
{
  nevot.tx.samples = nevot.audio.tick * nevot.tx.ad.sample_rate + 0.5;
  nevot.tx.bytes   = audio_samples_to_bytes(&nevot.tx.ad, nevot.tx.samples);
  nevot.rx.samples = nevot.audio.tick * nevot.rx.ad.sample_rate + 0.5;
  nevot.rx.bytes   = audio_samples_to_bytes(&nevot.rx.ad, nevot.rx.samples);
  ahw_set(0, AS_fillpoint, nevot.tx.samples);
} /* audio_chunk */
