/*
ring  --  ring buffer object

Copyright 1994 by Henning Schulzrinne; all rights reserved
*/

#include "audio_generic.h"
#include "ring.h"
#include "debug.h"
#include <stdlib.h>   /* malloc() */
#include <memory.h>   /* memset() */


extern struct RX_t RX;


/*
* Create a ring that and initialize it to silence.
* encoding I   audio encoding in buffer
*/
ring_t *ring_create(audio_encoding_t encoding)
{
  u_int32 size;
  ring_t *r;

  size = 1 << RING_BITS;
  r = (ring_t *)malloc(sizeof(ring_t) + size); 
  r->size        = size;
  r->mask        = (r->size - 1) & 0xfffffffe;
  r->next        = 0;
  r->last_sample = 0;
  r->member_last = -1;
  memset(r->member, 0, sizeof(r->member));
  audio_silence(r->buffer, r->size, encoding);
  return r;
} /* ring_create */

/*
* Call function 'action' on ring (starting at 'start', 'size' samples)
* as action(buf, count, arg1, arg2);
*/
int ring_action(ring_t *r, u_int32 start, u_int32 size,
  ring_action_t *action, void *arg1, void *arg2, int bpf)
{
  u_int32 first;    /* bytes left until end of buffer */

  /* convert from samples to bytes */
  if (bpf != 1) {
    size  *= bpf;
    start *= bpf;
  }

  start &= r->mask;
  first  = r->size - start;

  /* wrap around */
  if (first < size) {
           (*action)(&r->buffer[start], first,        arg1, arg2);
    return (*action)(&r->buffer[0],     size - first, arg1, arg2);
  }
  else {
    return (*action)(&r->buffer[start], size, arg1, arg2);
  } 
} /* ring_action */

/*
* Copy 'size' samples from 'src' buffer to ring buffer 'r' using copy routine
* 'cpy'. 'start': start sample within ring buffer (modulo 2^32); 
*/
void ring_cpy(ring_t *r, u_int32 start, u_int32 size,
  void (*cpy)(char *, char *, size_t), char *src, int bpf)
{
  u_int32 first;    /* bytes left until end of buffer */

  /* convert from samples to bytes */
  if (bpf != 1) {
    size  *= bpf;
    start *= bpf;
  }

  /* mask pointer and determine first chunk */
  start &= r->mask;
  first  = r->size - start;

  /* wrap around */
  if (first < size) {
    (*cpy)(&r->buffer[start], src,         first);
    (*cpy)(&r->buffer[0],     src + first, size - first);
  }
  else {
    (*cpy)(&r->buffer[start], src, size);
  }
} /* ring_cpy */

/*
* Return number of samples until buffer is depleted. Negative:
* time ring has been idle.
*/
int ring_left(ring_t *r)
{
  return r->last_sample - r->next;
} /* ring_left */

/*
* Clear ring, that is set entire ring to silence.
*/
void ring_clear(ring_t *r, audio_descr_t *ad)
{ 
  audio_silence(r->buffer, r->size, ad->encoding);
} /* ring_clear */
