
/*
session  --  functions for one conference/session

Copyright 1994 by Henning Schulzrinne; all rights reserved
*/

#include "nevot.h"
#include "udp.h"
#include "list.h"
#include "control.h"
#include "audio_generic.h"  /* */
#include "util.h"           /* random32() */
#include "member.h"
#include "des.h"            /* crypt_init() */
#include <stdlib.h>         /* free(), calloc() */
#include <errno.h>          /* EINVAL, etc. */
#include <string.h>         /* strcmp() */



typedef int (*int_func)();
static struct {
  char *name;                /* name of protocol */

  /* ... connect: tsap=  addres to connect to
   *              tsapl= address of local host -of relevance mainly for
   *                     multi homes hosts */
  int (*connect)(tsap_t *tsap, tsap_t *tsapl, Notify_func_input, 
		 Notify_client, void **);
  /* ... changes (multicast) ttl; return 0 if can't be changed */
  int (*set_ttl)(void *, int ttl);
  /* ... reads network data from socket */
  int (*read)(int sock, void *buf, int size, tsap_t *tsap);
  /* ... write network data to socket */
  int (*write)(void *, mbuf_t *b, tsap_t *);
  /* ... closes socket */
  int (*close)(void *);
} p[] = {
  {"UDP", 
    UDP_connect,
    UDP_ttl,
    UDP_read, UDP_write,
    UDP_close,
  },
#if TCP
  {"TCP",
    TCP_connect,
    ((int_func)0),
    TCP_read, TCP_write,
    TCP_close,
  }, 
#endif
};

static Audio_session_t last_session;

static Tcl_HashTable session_table;
static int session_table_exists;
extern struct TX_t TX;
extern struct RX_t RX;

/*
* Return 1 if session is multicast, 0 if not.  If s->local is zero, we
* can't determine it and return 0.
*/
int session_ismulticast(Audio_session_t s)
{
  return s->local && tsap_ismulticast(&s->local->ndst[0]);
} /* session_ismulticast */


/*
* Create the network sockets for member 'm' in a session.
* If session is multicast, don't do anything.
* Return 0 on success, -1 on error.
*/
Audio_errno_t session_open(Audio_session_t s, tsap_t *tsapp)
{
  int n;
  Audio_errno_t e = Audio_OK;       /* return status */
  extern int errno;

  for (n = 0; n < 2; n++) {
    tsap_setport(tsapp, s->sp.port + n);
    if (tsap_ismulticast(tsapp)) {
      tsap_cpy(&s->local->ndst[n], tsapp);
    }
    if ((*p[s->sp.net_proto].connect)(tsapp, &nevot.tsap[0],rx, 
				      (Notify_client)s+n,
				      n ? &s->net.control : &s->net.data) < 0){
      if      (errno == EINVAL)      e = Audio_EHNOMCAST;
      else if (errno == ENOPROTOOPT) e = Audio_ECNOMCAST;
      return e;
    }
  }
  /*
   * Now the sockets should be open so we can really set
   * the socket TTL, rather than just the s->ttl variable.
   */
  if (tsap_ismulticast(tsapp)) {
    Audio_SessionSetTTL(s, s->net.ttl);
  }
  return e;
} /* session_open */

/*
* Add a host named 'host' to session 's'.  Return -1 on error, 0
* otherwise.
*/
static Audio_errno_t session_add_host(Audio_session_t s, char *host)
{
  int n;
  Audio_errno_t e = Audio_OK;       /* return status */
  Audio_member_t m;
  tsap_t tsap;

  if (tsap_gethostbyname(host, &tsap, 1) == 1) {
    /*
     * Check if we are trying to add ourselves.
     */
    for (n = 0; n < nevot.tsap_count; n++) {
      if (tsap_equ_a(&nevot.tsap[n], &tsap)) {
        e = Audio_ESELF; 
        goto done;
      }
    }
    tsap_setport(&tsap, s->sp.port);
    /*
     * If unicast, add host to session's member list.
     * We can't add to a multicast session, thus, generate error
     * unless it's the same one we are already using.
     * For RTP, we don't know the remote site's src_id, thus we
     * set it to zero. Before creating a new site, we check whether
     * the same network address exists already and reuse that.
     */
    /* if equal to existing network connection, don't do anything */
    if (tsap_equ(&tsap, &s->local->ndst[0])) {
      printf("equal to existing network connection\n");
      goto done;
    }
    else if (session_ismulticast(s)) {
      e = Audio_EMIXED;
      goto done;
    }
    /* create multicast session: nothing to be done here */
    else if (tsap_ismulticast(&tsap)) {
    }
    /* add unicast member */
    else {
      if ((m = member_find(s, MEMBER_NODATA, &tsap,
         s->sp.rt_proto == R_vat ? tsap_ip4(&tsap) : 0)) == 0) {
        e = Audio_ERROR;
        goto done;
      }
    }
    /* Create network connection for session. */
    e = session_open(s, &tsap);
  }
  else {
    e = Audio_ENOHOST;
    goto done;
  }

done:
  tsap_free(&tsap);
  return e;
} /* session_add_host */


/*
* Locate session based on textual session identifier assigned
* in Audio_SessionAlloc().
*/
Audio_session_t Audio_SessionFind(char *id)
{
  Tcl_HashEntry *entry;

  if (!session_table_exists) return (Audio_session_t)0;
  entry = Tcl_FindHashEntry(&session_table, id);
  if (entry == NULL) return (Audio_session_t)0;
  return (Audio_session_t)Tcl_GetHashValue(entry);
} /* Audio_SessionFind */


/*
* Return textual session identifier of session 's'.
*/
char *Audio_SessionId(Audio_session_t s)
{
  return s->name;
} /* Audio_SessionId */

/*
* Delete session descriptor 's'.  Notify conference controller with
* 'closed' message.  Return Audio_OK on success; currently no error
* value.
*/
Audio_errno_t Audio_SessionDelete(Audio_session_t s)
{
  Audio_member_t m, m_next;
  Tcl_HashEntry *entry;
  sdes_t *sdes, *sdes_next;

  /* Disable new members for this session. */
  s->no_more = TRUE;

  /* Remove periodic function. */
  notify_set_periodic_func((Notify_client)s, tx_periodic, 0);
  tx_packet_session(s, 0, P_bye);

  /* Delete all members of this session. */
  for (m = s->first; m; m = m_next) {
    m_next = m->next;
    Audio_MemberDelete(m, "Session deleted");
  }

  /*
   * Close network connection (must be after Audio_MemberDelete() as
   * that sends BYE packets.
   */
  if (s->net.close) {
    (*s->net.close)(s->net.control);
    (*s->net.close)(s->net.data);
  }

  /* Free the ring buffer if nobody else is sharing it. */
  if (s->ring) {
    Audio_session_t stp;

    ALL_SESSIONS(stp) {
      if (stp->ring == s->ring && s != stp) break;
    }
    if (!stp) free(s->ring);
  }

  /* Free any left-over attached transmit mbufs. */
  m_freem(s->b_head, "sd");

  /* Delete hash table. */
  Tcl_DeleteHashTable(&s->member_table);

  /* Delete SDES information. */
  for (sdes = s->sdes; sdes; sdes = sdes_next) {
    if (sdes->value) free(sdes->value);
    sdes_next = sdes->next;
    free(sdes);
  }

  /* Unlink session descriptor from global list. */
  if (s->next) (s->next)->prev = s->prev;
  if (s->prev) (s->prev)->next = s->next;
  if (last_session == s) last_session = s->prev;
  if (nevot.first == s) nevot.first = s->next;
  if (nevot.first == 0) {  /* free last and only session */
    nevot.audio.talking = FALSE;
    audio_close();
  }
  (*nevot.event)(s, 0, Audio_closed, 0);

  if (!session_table_exists) return 0;
  entry = Tcl_FindHashEntry(&session_table, s->name);
  if (s->name) free(s->name);
  if (s->key)  free(s->key);
  Tcl_DeleteHashEntry(entry);
  free(s);
  return Audio_OK;
} /* Audio_SessionDelete */

/*
* Allocate the basic session storage, but don't connect anything yet.
* Return -1 on error (likely bug), 0 on success.
* Marked by 'members' == -1.
*/
Audio_errno_t Audio_SessionAlloc(char *id, Audio_sp_t *sp,
  Audio_session_t *stp)
{
  Audio_session_t s;
  int new;
  Tcl_HashEntry *entry;

  s = (Audio_session_t)calloc(1, sizeof(struct Audio_session_t));
  if (!s) return Audio_ERROR;
  if (!nevot.first) { /* this is the first session */
    nevot.first = last_session = s;
    /* create ring */
    s->ring = ring_create(nevot.rx.ad.encoding);
  }
  else {
    s->next = (Audio_session_t)0;
    s->prev = last_session;
    last_session->next = s;
    last_session = s;

    /* If end system, all use the same ring buffer; gateways have own
     * ring for each session.
     */
    s->ring = (nevot.role == R_end) ? 
               nevot.first->ring : 
               ring_create(nevot.rx.ad.encoding);
  }
  s->members    = -1;
  s->talk       = FALSE;
  s->seq        = random32((int)&s->seq);
  s->initial_ts = random32((int)&s->initial_ts);
  /* rest is zero */

  /* default properties */
  if (sp) s->sp = *sp;

  /* default values for delay adaptation */
  if (s->sp.delay.max <= 0.)        s->sp.delay.max = 5.;
  if (s->sp.delay.initial <= 0.)    s->sp.delay.initial = 0.05; 
  if (s->sp.delay.multiplier <= 0.) s->sp.delay.multiplier = 4;
  s->sp.delay.dvar_x = 0.1;
  s->sp.delay.dvar_y = 1 - s->sp.delay.dvar_x;

  /* enter into hash table */
  s->name = strdup(id);
  if (!session_table_exists) {
    Tcl_InitHashTable(&session_table, TCL_STRING_KEYS);
    session_table_exists = 1;
  }
  entry = Tcl_CreateHashEntry(&session_table, id, &new);
  if (!new) return Audio_EDUP;
  Tcl_SetHashValue(entry, s);

  *stp = s;
  return Audio_OK;
} /* Audio_SessionAlloc */

/*
* For session-oriented sessions, a listen socket is created and bound.
* If members are present, they are connected as well at this point.
* If session exists, the members are added.
* Return Audio_OK on success, Audio_* on error.
*/
Audio_errno_t Audio_SessionOpen(Audio_session_t s, char *host)
{
  Audio_sp_t *sp = &s->sp; /* session characteristics */
  Audio_errno_t e;
  sdes_t *sdes;

  /* Disallow odd port numbers. */
  if (sp->rt_proto == R_rtp && (sp->port & 1)) return Audio_EODDPORT;

  if (!s->net.read) {
    /* Initialize RT protocol function pointers. */
    if      (sp->rt_proto == R_vat) vat_init(&s->rt);
    else if (sp->rt_proto == R_rtp) rtp_init(&s->rt);
    else return Audio_EPROTONOSUPPORT;

    /* Initialize outgoing audio encoding. */
    Audio_SessionSetAudio(s, &s->audio, s->local_audio_channel, s->packetization);

    /* Add hash table. */
    Tcl_InitHashTable(&s->member_table, TCL_ONE_WORD_KEYS);

    /* 
     * Add ourselves (local site).
     */
    s->local = member_find(s, MEMBER_NODATA, &nevot.tsap[0],
     sp->rt_proto == R_rtp ? random32((int)s) : tsap_ip4(nevot.tsap));
  }

  /* Notify others that the session has been opened. */
  (*nevot.event)(s, 0, Audio_opened, host);

  /* Add SDES information to local site. */
  for (sdes = s->sdes; sdes; sdes = sdes->next) {
    member_sdes(s->local, sdes->type, sdes->value, sdes->len);
  }

  /* Only if not established already. */
  if (!s->net.read) {
    /* Set up network functions; before session_add_host to get ttl to work. */
    s->net.read    = p[sp->net_proto].read;
    s->net.write   = p[sp->net_proto].write;
    s->net.set_ttl = p[sp->net_proto].set_ttl;
    s->net.close   = p[sp->net_proto].close;
  }

  /* Network interface: data and control. */
  if ((e = session_add_host(s, host)) < 0) return e;

  /* Add periodic callback to send status messages (immediately). */
  tx_periodic((Notify_client)s);

  /* Enable listening. */
  Audio_SessionListen(s, TRUE);

  /* Allow adding members now, after session is all set up. */
  s->no_more = FALSE;

  return Audio_OK;
} /* Audio_SessionOpen */

/*
* Set/modify outgoing SDES value. Needs to modify s->local rather
* than s->sp since s->sp.sdes is allocated by the caller of
* Audio_SessionOpen().
* Return 0 on success, Audio_EINVAL on error.
*/
Audio_errno_t Audio_SessionSetSDES(Audio_session_t s, const SDES_t type,
  const char *data)
{
  struct sdes *sdes, *sdes_last = 0;

  for (sdes = s->sdes; sdes; sdes = sdes->next) {
    if (sdes->type == type) break;
    sdes_last = sdes;
  }

  /* Create a new SDES entry. */
  if (!sdes) {
    sdes = (struct sdes *)malloc(sizeof(struct sdes));
    if (sdes_last) sdes_last->next = sdes;
    else s->sdes = sdes;
    sdes->next  = 0;
    sdes->value = 0;
    sdes->len   = 0;
    sdes->type  = type;
  }

  if (sdes->value) {
    free(sdes->value);
    sdes->value = NULL;
  }
  if (strlen(data) > 0) {
    sdes->value = strdup(data);
    sdes->len   = strlen(data);
  }
  if (s->local) member_sdes(s->local, sdes->type, sdes->value, sdes->len);

  return Audio_OK;
} /* Audio_SetSDES */

/*
* Set multicast ttl (time-to-live) value.  Ignored for unicast.  Store
* in 's' structure even if currently no setting routine (for later
* use).  Return Audio_OK on success, Audio_EINVAL if ttl value is out
* of range.
*/
Audio_errno_t Audio_SessionSetTTL(Audio_session_t s, int ttl)
{
  if (ttl < 0 || ttl > 255) return Audio_EINVAL;
  s->net.ttl = ttl;
  if (s->net.set_ttl) {
    (*s->net.set_ttl)(s->net.data,    s->net.ttl);
    (*s->net.set_ttl)(s->net.control, s->net.ttl);
  }
  return Audio_OK;
} /* Audio_SessionSetTTL */

/*
* Allocates memory block for key schedule.
* Return 0 if key is weak or other error.
*/
static void *crypt_init(char *str)
{
  des_key_schedule *schedule;
  des_cblock key;
  char algorithm[16];

  des_check_key = 1;
  string_DES_key(str, (u_int8 *)&key, algorithm);
  schedule = (des_key_schedule *)malloc(sizeof(des_key_schedule));
  des_set_key(&key, *schedule);  
  return (void *)schedule;  
} /* crypt_init */

/*
* Set outgoing encryption key. A value of NULL disables encryption.
* Return Audio_OK on success, Audio_EINVAL if key is invalid.
*/
Audio_errno_t Audio_SessionSetKey(Audio_session_t s, char *key)
{
  s->encrypt = (key != NULL && *key != '\0');
  if (s->encrypt) s->key = crypt_init(key);

  return Audio_OK;
} /* Audio_SessionSetKey */

/*
* Set duplex/echo suppression mode.
* Return Audio_OK on success, Audio_EINVAL if invalid parameter.
*/
Audio_errno_t Audio_SessionSetDuplex(Audio_session_t s,
  Audio_duplex_t d)
{
  if (d < Audio_duplex || d > Audio_local_priority) return Audio_EINVAL;
  s->m.net_priority = (d == Audio_net_priority);
  s->m.local_priority = (d == Audio_local_priority);
  return Audio_OK;
} /* Audio_SessionSetDuplex */

/*
* Set outgoing audio format (encoding, sampling rate, channels), the
* stereo channel incoming audio is played on (default: 1) and the
* number of seconds of audio per outgoing network packets.  Adjust
* local (input) audio device accordingly, but do not adjust
* packetization to allow several different sessions with different
* packetization times.  If 'packetization' is zero, leave unchanged. 
* Does not enable audio device.  Return Audio_OK on success,
* Audio_EINVAL if invalid parameter, Audio_ERATE if invalid sampling
* rate, Audio_ECHANNELS if invalid channel count, Audio_EENC if
* invalid encoding. If 'ad' is zero, don't change.
*/
Audio_errno_t Audio_SessionSetAudio(Audio_session_t s,
  audio_descr_t *ad, int audio_channel, double chunk)
{
  Audio_errno_t status = Audio_OK;

  /* check chunk for validity */
  if (chunk != 0 && (chunk < 0. || chunk > MAX_PACKETIZATION || chunk < nevot.audio.tick))
    return Audio_EINVAL;

  /*
   * Change audio input parameters and update dependent parameters.
   * We don't exit immediately if this fails, as it may be due to
   * a temporarily unavailable audio device.
   */
  if (ad) {
    status = audio_change(0, ad);
    s->audio = *ad;
    s->control.bw = audio_seconds_to_bytes(ad, 1.) * 0.05;
    ring_clear(s->ring, ad);
  }

  /* Change outgoing packetization interval. */
  s->packetization = chunk;
  s->bpm = chunk / nevot.audio.tick + 0.5;

  /* Set up RTP/vat for new outgoing audio encoding. */
  if (s->rt.change_audio && (*s->rt.change_audio)(&s->rt, &s->audio) < 0)
    return Audio_EENC;

  return status;
} /* Audio_SessionSetAudio */

/*
* Set session talk variable. Return Audio_OK on success,
* Audio_EAUDIOOPEN or Audio_EAUDIOBUSY on failure.
* Audio only needs to be turned on if ring is active or we are talking.
*/
Audio_errno_t Audio_SessionTalk(Audio_session_t s, const int talk)
{
  Audio_errno_t e;
  s->talk = talk;

  /* check if any session is talking */
  nevot.audio.talk = 0;
  ALL_SESSIONS(s) {
    if (s->talk) {
      nevot.audio.talk = 1;
      break;
    }
  }

  /* activate/deactivate audio device */
  if (nevot.audio.talk) {
    if ((e = audio_open()) < 0) {
      s->talk = 0;
      return e;
    }
  }
  else {
    audio_vu(0, 0.);
    audio_close_idle();
  }
  return Audio_OK;
} /* Audio_SessionTalk */


/*
* Enable and disable listening for whole session.
* Return Audio_OK.
*/
Audio_errno_t Audio_SessionListen(Audio_session_t s, const int listen)
{
  s->listen = listen;
  (*nevot.event)(s, 0, Audio_receive, s->listen);
  if (!listen) audio_vu(1, 0.);
  return Audio_OK;
} /* Audio_SessionListen */
