/*
* Device-independent audio interface, using ahw_...() routines.
*/
#include <tcl.h>
#include <string.h>   /* strcmp(), strchr() */
#include <errno.h>
#include <stdlib.h>   /* atoi() */
#include "ahw.h"
#include "tkaudio.h"

static char rcsid[] = "$Id: tkaudio_trace.c,v 1.1.1.1 1998/04/21 17:05:50 sho Exp $";

static struct {
  char *name;
  audio_port_t port[2];
} list[] = {
  {"microphone", {AP_in_microphone, 0}},
  {"line",       {AP_in_line,       AP_out_line}},
  {"headphone",  {0,                AP_out_headphone}},
  {"speaker",    {0,                AP_out_speaker}},
  {"digital",    {AP_in_digital,    AP_out_digital}},
  {NULL,         {0,                0}}
};

extern char *sys_errlist[];


/*
* Variable trace handler for ports and gain.
* Disable if called indirectly from tkaudio_poll.
*/
char *tkaudio_gain_trace(ClientData client, Tcl_Interp *interp,
  char *name1, char *name2, int flags)
{
  char *v = Tcl_GetVar2(interp, name1, name2, flags & TCL_GLOBAL_ONLY);
  int status = 0;
  int value;
  int pr = (long)client;
  audio_state_t a;
  char str[40];

  value = atoi(v);
  if (value < 0) {
    value = 0;
    Tcl_SetVar2(interp, name1, name2, "0", flags & TCL_GLOBAL_ONLY);
  }
  else if (value > 100) {
    value = 100;
    Tcl_SetVar2(interp, name1, name2, "100", flags & TCL_GLOBAL_ONLY);
  }
  ahw_track(0);
  status = ahw_set(pr, AS_gain, (value * 255) / 100);
  /* 
   * Since the recording and play gain may be quantized at less than
   * 1 unit, we need to read the actual value and set the Tcl
   * variables accordingly.
   */
  if (ahw_get(&a) < 0) return sys_errlist[errno];

  if (pr == 0) {
    sprintf(str, "%1.0f", (a.record.gain[0] * 100.) / 255);
    Tcl_SetVar2(interp, "record", "gain", str, TCL_GLOBAL_ONLY);
  }
  else if (pr == 1) {
    sprintf(str, "%1.0f", (a.play.gain[0] * 100.) / 255);
    Tcl_SetVar2(interp, "play", "gain", str, TCL_GLOBAL_ONLY);
  }
  else if (pr == 2) {
    sprintf(str, "%1.0f", (a.monitor_gain[0] * 100.) / 255);
    Tcl_SetVar2(interp, "monitor", "gain", str, TCL_GLOBAL_ONLY);
  }

  ahw_track(tkaudio_poll);
  if (status < 0) return sys_errlist[errno];
  return NULL;
} /* tkaudio_gain_trace */

/*
* Tcl variable trace for port.
*/
char *tkaudio_port_trace(ClientData client, Tcl_Interp *interp,
  char *name1, char *name2, int flags)
{
  char *v = Tcl_GetVar2(interp, name1, name2, flags & TCL_GLOBAL_ONLY);
  int i;
  int p = -1;
  int pr = (long)client;  /* play or record */
  int status;

  /* search name in port list */
  for (i = 0; list[i].name; i++) {
    if (list[i].port[pr] && (strcmp(v, list[i].name) == 0)) {
      p = list[i].port[pr];
      break;
    }
  }
  if (p > 0) {
    ahw_track(0);
    status = ahw_set(pr, AS_port, p);
    ahw_track(tkaudio_poll);
  }

  if (status < 0) return sys_errlist[errno];
  return NULL;
} /* tkaudio_port_trace */

/*
* Return list of supported ports: {{microphone line} {speaker
*   headphone}}
* No arguments.
*/
int tkaudio_ports(ClientData client, Tcl_Interp *interp, int argc,
  char *argv[])
{
  int i, pr;
  audio_state_t a;

  if (ahw_get(&a) < 0) {
    strcpy(interp->result, sys_errlist[errno]);
    return TCL_ERROR;
  }
  for (pr = 0; pr <= 1; pr++) {
    Tcl_AppendResult(interp, "{", NULL);
    for (i = 0; list[i].name; i++) {
      if (list[i].port[pr] && (ahw_set(pr, AS_port, list[i].port[pr])) >= 0) {
        Tcl_AppendElement(interp, list[i].name);
      }
    }
    Tcl_AppendResult(interp, "} ", NULL);
  } 

  /* restore prior state */
  ahw_set(0, AS_port, a.record.port);
  ahw_set(1, AS_port, a.play.port);

  return TCL_OK;
} /* tkaudio_ports */

/*
* Get current values of the audio parameters and set Tcl variables
* accordingly. 
* changes=0: read initial values into Tcl variables.
* changes=1: just read changes.
*/
static void tkaudio_get(ClientData changes)
{
  int i, k;
  audio_state_t c;
  static audio_state_t a;
  char str[40];
  extern Tcl_Interp *Interp;

  if (ahw_get(&c) < 0) return;

  /* gain for each channel */
  for (i = 0; i < 1; i++) {
    if (c.monitor_gain[i] != a.monitor_gain[i] || !changes) {
      sprintf(str, "%1.0f", (c.monitor_gain[i] * 100.) / 255);
      Tcl_SetVar2(Interp, "monitor", "gain", str, TCL_GLOBAL_ONLY);
    }
  }

  /* for input and output */
  for (i = 0; i < 2; i++) {
    audio_ws_t *new, *old;
    char *dir;

    dir = (i == 0) ? "play"  : "record";
    new = (i == 0) ? &c.play : &c.record;
    old = (i == 0) ? &a.play : &a.record;

    /* audio port */
    if (new->port != old->port || !changes) {
      for (k = 0; list[k].name; k++) {
        if (new->port & list[k].port[1-i]) {
          Tcl_SetVar2(Interp, dir, "port", list[k].name, TCL_GLOBAL_ONLY);
          break;
        }
      }
      if (!list[k].name) {
        Tcl_SetVar2(Interp, dir, "port", "invalid", TCL_GLOBAL_ONLY);
      }
    }

    /* gain */
    if (new->gain[0] != old->gain[0] || !changes) {
      sprintf(str, "%1.0f", (new->gain[0] * 100.) / 255);
      Tcl_SetVar2(Interp, dir, "gain", str, TCL_GLOBAL_ONLY);
    }

    if (new->samples != old->samples || !changes) {
      sprintf(str, "%d", new->samples);
      Tcl_SetVar2(Interp, dir, "samples", str, TCL_GLOBAL_ONLY);
    }

    if (new->active != old->active || !changes) {
      sprintf(str, "%d", new->active);
      Tcl_SetVar2(Interp, dir, "active", str, TCL_GLOBAL_ONLY);
    }

    if (new->errors != old->errors || !changes) {
      sprintf(str, "%d", new->errors);
      Tcl_SetVar2(Interp, dir, "errors", str, TCL_GLOBAL_ONLY);
    }

    if (new->ad.encoding != old->ad.encoding || !changes) {
      char *s;
      extern char *audio_ntoa(audio_descr_t *);

      strcpy(str, audio_ntoa(&new->ad));
      /* keep only encoding, discard sample_rate, etc. */
      s = strchr(str, ' ');
      if (s) *s = '\0'; 
      Tcl_SetVar2(Interp, dir, "encoding", str, TCL_GLOBAL_ONLY);
    }

    if (new->ad.channels != old->ad.channels || !changes) {
      sprintf(str, "%d", new->ad.channels);
      Tcl_SetVar2(Interp, dir, "channels", str, TCL_GLOBAL_ONLY);
    }

    if (new->ad.sample_rate != old->ad.sample_rate || !changes) {
      sprintf(str, "%d", new->ad.sample_rate);
      Tcl_SetVar2(Interp, dir, "sample_rate", str, TCL_GLOBAL_ONLY);
    }
  }

  a = c;
} /* tkaudio_get */


/*
* Signal handler invoked if audio parameter changes.  Since this may
* trigger lots of Tcl and Tk code, and doing this from within a signal
* handler leads to X crashing and other random effects, we only
* schedule a delayed call from the signal, hoping that that is safe.
*/
void tkaudio_poll(int sig)
{ 
   Tcl_CreateTimerHandler(50, tkaudio_get, (ClientData)1); 
} /* tkaudio_poll */


int tkaudio_state(ClientData client, Tcl_Interp *interp, int argc,
  char *argv[])
{
  tkaudio_get((ClientData)1);
  return TCL_OK;
} /* tkaudio_state */

/*
* Set up a periodic timer that checks the audio device for any
* changes. For systems that signal changes, ahw_get() is only
* called if the ahw_changed flag is set.
*/
void tkaudio_timer(ClientData client)
{
  /* get changed or initial values */
  if (ahw_changed || client == 0) tkaudio_get((ClientData)ahw_changed);
  /* schedule next timer */
  Tcl_CreateTimerHandler(2000, tkaudio_timer, (ClientData)1);
} /* tkaudio_timer */
