/*
* Search in alias file for any matching aliases.
* Lines beginning with white space are treated as continuation lines;
* lines beginning with # are comments. Aliases may recurse.
* Returns list of space-separated aliases.
*/

#include <search.h>
#include <string.h>
#include <ctype.h>
#include <stdio.h>
#include <malloc.h>
#include <stdlib.h>

/*
* Read alias file containing one name per line. Return space-separated
* list, dynamically allocated.
*/
static char *alias_file(char *filename)
{
  FILE *f;  
  int len = 0;
  char *list = 0;
  char *word;   /* start of word */
  char *s;
  char line[1024];

  if (!(f = fopen(filename, "r"))) {
    perror(filename);
    return NULL;
  }
  while (fgets(line, sizeof(line), f)) {
    /* Strip leading and trailing white space. */
    for (word = line; *word && isspace(*word); word++) ;
    for (s = word; *s && !isspace(*s); s++) ;
    *s = '\0';
    list = list ? realloc(list, len + strlen(word) + 1) :
                  malloc(strlen(word)+1);
    sprintf(&list[len], "%s\n", word);
    len += strlen(word)+1;  /* include \n */
  }
  fclose(f);
  return list;
} /* alias_file */


char *aliases(char *aliasfile, char *name, char *list)
{
  FILE *f;
  char line[1024];
  char word[128]; /* safe since LOGNAME_MAX is 8 */
  char *s, *d, *name_stripped;
  int len = 0;    /* length of list */
  int more = 0;   /* wait for more continuation lines */

  /* Ignore empty words. */
  if (!name || !name[0]) return 0;

  /*
   * Remove all but numbers and digits from name to allow handling of
   * phone numbers such as +49-30-254-99-200.
   */
  name_stripped = malloc(strlen(name) + 1);
  for (s = name, d = name_stripped; *s; s++) {
    if (!isspace(*s) && !ispunct(*s)) *d++ = *s;
  }
  *d = '\0';
  name = name_stripped;

  if (list) len = strlen(list);
  if (!(f = fopen(aliasfile, "r"))) {
    perror(aliasfile);
    return NULL;
  }
  while (fgets(line, sizeof(line), f)) {
    if (!line[0] || line[0] == '#' || line[0] == '\n') {
      more = 0;
      continue;
    }
    /* Continuation line. */
    if (more && isspace(line[0])) {
      s = line;
    }
    /* 
	 * Alias entry:  Line starting with text (typically, letters but
	 * could also be a phone number.
     */
    else if (!isspace(line[0])) {
      more = 0;
      for (s = line, d = word; *s && *s != ':'; s++) {
        *d++ = *s; 
      }
      *d = '\0';
      if (*s) s++; /* skip : */
    }

    /*
     * Get comma-separated list of aliases after colon.
     */
    if (more || !strcasecmp(name, word)) {
      /* Check if alias is of form :include:file */
      for (; *s && isspace(*s); s++) ;
      if (!strncmp(s, ":include:", strlen(":include:"))) {
        for (d = s; *d && !isspace(*d); d++) ;
        *d = '\0';  /* remove trailing white space */
        list = alias_file(s + strlen(":include:"));
      }
      else {
        /* Otherwise, this is a list of names (users or aliases). */
        more = 1;
        while (*s) {
          for (; *s && isspace(*s); s++) ;
          /* Get one expansion. */
          for (d = word; *s && *s != ','; s++) {
            if (!isspace(*s)) *d++ = *s;
          }
          *d = '\0';
          if (*s == ',') s++;  /* skip comma */
          /* Recurse. */
          if (word[0]) {
            d = aliases(aliasfile, word, 0);
            if (!d) {  /* no expansion found, thus add \n */
              strcat(word, "\n");
              d = word;
            }
            list = list ? realloc(list, len + strlen(d) + 1) :
                          malloc(strlen(d) + 1);
            strcpy(&list[len], d);
            len += strlen(d);
            if (d != word) free(d); /* d was an alias. */
          }
        }
      }
    }
  }
  fclose(f);
  free(name);  /* remove stripped name */
  return list;
} /* aliases */


static void usage(char *pname)
{
  fprintf(stderr, "Usage: %s [-a aliasfile] name\n", pname);
  exit(1);
} /* usage */


int main(int argc, char *argv[])
{
  extern char *optarg;
  extern int optind;
  char *list;
  char *aliasfile = "/etc/aliases";
  int c;

  while ((c = getopt(argc, argv, "a:")) != EOF) {
    switch (c) {
    case 'a':
      aliasfile = optarg;
      break;
    case '?':
      usage(argv[0]);
      break;
    }
  }
  /* Not enough arguments. */
  if (optind >= argc) {
    usage(argv[0]);
  }

  /* Search for aliases in system alias file. */
  list = aliases(aliasfile, argv[optind], 0);
  if (list) {
    fputs(list, stdout);
    free(list);
  }
  return 0;
} /* main */
