/***********************************************************************************
 *
 * procedures needed to communicate with isc
 * - start isc
 * - event handler for isc_socket
 * - send messages to isc
 *
 * partially based on cd by Christian Zahl
 *
 * Copyright 1997 by Stefan Hoffmann
 * GMD Fokus 
 *
 ***********************************************************************************/



#include <stdlib.h>
#include <string.h>
#include <pwd.h>
#include <signal.h>
#include <syslog.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <unistd.h>
#include "debug.h"
#include "defines.h"
#include "list.h"
#include "misc.h"
#include "notify.h"
#include "multimer.h"
#include "timer.h"
#include "reply_codes.h"
#include "response.h"
#include "udp.h"



extern char	myAddr[17];
extern int	isc_socket;
extern inv_list_t *invitation_list;
extern int	isc_connections;
extern char     isc_bin_path[255];
extern char     isc_audio_path[255];
extern char     StartIscService[255];


/* isc_socket_event *****************************************************************
 *
 *  Handler for messages from isc via local multicast
 *  invoked when local multicast socket becomes readable
 *
 ***********************************************************************************/
void isc_socket_event (Notify_client client, int fd)
{
  static char         func[] = "isc_socket_event";
  char                *msg; /* message received from isc */
  char                pmm_header[MAX_CALL_ID_LENGTH+6];
  int                 len;
  char                resp_code_str[4];
  int                 resp_code;
  int                 sec;
  char                *p;
  inv_list_t          *inv;
  struct timeval      time;


  DEBUG (2, syslog (LOG_DEBUG, "%s",func);)

  /* get some memory for holding the message */
  if ((msg = (char *)malloc (MSG_BUF_SIZE)) == NULL) 
    return;
 
  if ((len = recvfrom(isc_socket , msg, MSG_BUF_SIZE, 0, NULL, 0)) == -1) {
    free(msg);
    return;
  }

  /* terminate msg */
  msg[len] = '\0';
  
  /***********************************************************************************
   * check if msg was a "+ call_id" message which indicates, that isc was started
   ***********************************************************************************/
  if ((msg[0] == '+') && (len > 3)) {
    /* check call_id */
    if ((inv = inv_list_search(msg + 2)) != NULL) {
      /* kill T_start_isc and T_isc timer */
      timer_set(NULL, (Notify_func)" ", (Notify_client)inv, 1);
      if (send_to_isc(inv, inv->request) < 0) {
      /***********************************************************************************
       * sending the request to isc failed; handle like no response from isc was received 
       * if there are still locations to query, otherwise send error reply 
       ***********************************************************************************/
        free(msg);
        if (inv->locations > 1)
          isc_request_timeout((Notify_client)inv);
        else {
          inv->locations = 0;
          reply (inv, SERVER_INTERNAL_ERROR, "Server Internal Error"); 
        }
        return;
      }
      /* start T_isc timer */
      time.tv_sec = 2;
      time.tv_usec = 0;
      timer_set(&time, (Notify_func)start_isc_timeout, (Notify_client)inv, 1); 
    } 
    else {
      free(msg);
      return;
    }
  }

  /***********************************************************************************
   * check if msg was a SIP msg 
   ***********************************************************************************/
  p = msg;
  if ((p = get_field (p, pmm_header, sizeof(pmm_header)-1)) == NULL) {
    free(msg);
    return;
  }
  
  /* replace all "_" characters by "/" in pmm_header and get call_id */
  while ((p = strchr(pmm_header, '_')) != NULL)
    *p = '/';
 
  if ((p = strstr(pmm_header, "/sip/")) == NULL) {
    free (msg);
    return;
  }   
  *p = '\0';
  /* call-id is now stored in pmm_header */
  /* search for invitation list entry specified by call-id */
  if ((inv = inv_list_search(pmm_header)) == NULL) {
    free(msg);
    return;
  }

  /* skip pmm-header */
  memmove(msg, strstr(msg, " ")+1, strlen(msg)-(msg-strstr(msg, " ")+1));

  /***********************************************************************************
   * check if a response was received with the correct SIP version
   ***********************************************************************************/
  if ((str_search(msg, SIP_VERSION)) != msg) {
    free(msg);
    return;
  }

  /* we received a response msg */
  /* replace all double TAB by CRLF */
  while ((p = strstr(msg, "\t\t")) != NULL) {
    *p++ = '\r'; 
    *p = '\n';
  }

  /* replace all single TAB by LF */
  while ((p = strstr(msg, "\t")) != NULL) {
    *p = '\n';
  }
  
  /***********************************************************************************
   * get and check response code 
   ***********************************************************************************/
  if (get_field(msg+strlen(SIP_VERSION)+1, resp_code_str, sizeof(resp_code_str)) == NULL) {
    free(msg);
    return;
  }
  resp_code = atoi(resp_code_str);

  if (resp_code == 100) {
    /* isc received the request, so kill T_start_isc or T_isc timer */
    timer_set(NULL, (Notify_func)" ", (Notify_client)inv, 1);
  
    /* start T_isc_request timer for receiving a final response from isc */
    sec = ISC_TIMEOUT;
    time.tv_sec = sec; 
    time.tv_usec = 0; 
    timer_set(&time, (Notify_func)isc_request_timeout, (Notify_client)inv, 1); 
  }

  if (get_response_priority(resp_code) >= 20) {
    /***********************************************************************************
     * a final response was received, kill T_isc_req timer, close isc_socket 
     * if there are no other connections to isc
     ***********************************************************************************/
    timer_set(NULL, (Notify_func)isc_request_timeout, (Notify_client)inv, 1);
    if (--isc_connections <= 0) {
      notify_set_input_func(0, NOTIFY_FUNC_INPUT_NULL, isc_socket);
      time.tv_sec = 1;  
      time.tv_usec = 0;
      timer_set(&time, (Notify_func)close_isc_socket, (Notify_client)0, 1);
    }
  }
  
  process_response(msg, myAddr);
} /* isc_socket_event */



/* send_to_isc **********************************************************************
 *
 *  send a SIP message via local multicast to isc
 *  - generates pmm-header based on the call-id
 *
 *  in:
 *  - inv: pointer to entry in invitation_list 
 *  - msg: pointer to message which should be sent to isc
 *  
 *  returns:
 *  - bytes sent
 *
 ***********************************************************************************/
int send_to_isc (inv_list_t *inv, char *msg)
{
  static char     func[] = "send_to_isc";
  char            call_id_tmp[strlen(inv->call_id)];
  char            msg_tmp[strlen(msg)];
  char            *c;


  DEBUG (2, syslog (LOG_DEBUG, "%s", func);)

  /***********************************************************************************
   * create pmm header 
   ***********************************************************************************/
  strcpy(call_id_tmp, inv->call_id);
  /* replace all "/" characters by "_" in call_id to avoid conflicts with the pmm header */
  while ((c = strchr(call_id_tmp, '/')) != NULL)
    *c = '_';
 
  strcpy(msg_tmp, msg);
  /***********************************************************************************
   * replace all CR and LF to enable isc to read the whole SIP message and not to stop 
   * after receiving a CR or LF
   ***********************************************************************************/
  /* replace all TAB by SPACE */
  while ((c = (char *)strchr(msg_tmp, '\t')) != NULL) 
    *c = ' ';
    
  /* replace all CR by TAB */
  while ((c = (char *)strchr(msg_tmp, '\r')) != NULL) 
    *c = '\t';
    
  /* replace all LF by TAB */
  while ((c = (char *)strchr(msg_tmp, '\n')) != NULL) 
    *c = '\t';
    
  return UDP_send(LOCAL_MCAST_ADDR, LOCAL_MCAST_PORT, isc_socket, 3, call_id_tmp, "/sip/0 ", msg_tmp);
} /* send_to_isc */



/* start_isc *************************************************************************
 *
 *  starts the conference controller isc 
 *  invoked when isc_timeout expires
 *  - creates child process which executes start_isc script 
 *
 *  in:
 *  - passwd structure of invited user
 *  - call_id
 *
 *  returns:
 *  - 0, ok
 *  - -1, faliure
 *
 ***********************************************************************************/
int start_isc (struct passwd *userEntry, int call_id)
{
  static char     func[] = "start_isc";
  char            buf[4*1024];
  char            *envv[10];
  pid_t           pid;
  char            *p;
  int             i;
  char            string[255];
  char            string2[255];
  struct passwd* info;

  DEBUG (2, syslog (LOG_DEBUG, "%s", func);)

  /* start isc in the background */
  signal (SIGCHLD, SIG_DFL);
  pid = fork ();
 
  if (pid == -1) 
    return -1;
  
  /* we are the child, so start the start_isc script */
  if (pid == 0) {
    if (setgid (userEntry->pw_gid))
      DEBUG (1, syslog (LOG_WARNING, "%s: warning: cannot setgid: %m", func);)
    if (setuid (userEntry->pw_uid))
      DEBUG (1, syslog (LOG_WARNING, "%s: warning: cannot setuid: %m", func);)
    if (chdir (userEntry->pw_dir))
      DEBUG (1, syslog (LOG_WARNING, "%s: warning: cannot chdir: %m", func);)  
    closelog ();

    if (*StartIscService != NULL) {
      i = 0;
      envv[i++] = p = buf;
      sprintf (p, "HOME=%s", userEntry->pw_dir);
      envv[i++] = p += strlen (p) +1;
      sprintf (p, "USER=%s", userEntry->pw_name);
      envv[i++] = p += strlen (p) +1;
      sprintf (p, "DISPLAY=%s", ":0.0");   
      envv[i++] = p += strlen (p) +1;   
      sprintf (p, "ARCH=%s", ARCH);
      envv[i++] = NULL;
      
      execle (StartIscService, StartIscService, call_id, NULL, envv);   
    }
    else {
      putenv("DISPLAY=:0.0");
      p = buf;
      sprintf(p, "HOME=%s", userEntry->pw_dir); 
      putenv(p); 
      p += strlen (p) +1;
      sprintf(p, "USER=%s", userEntry->pw_name);   
      putenv(p);   
      if (*isc_audio_path != NULL) {
	p += strlen (p) +1;
	sprintf(p, "ISC_AUDIO=%s", isc_audio_path);   
	putenv(p);
      }
      
      p += strlen (p) +1;
      sprintf(p, "%s %s%s %s %s", "exec", isc_bin_path, " -i -C", call_id, "> /dev/null &");   
      system (p);  
    }
    exit (0);
  
  }

  waitpid (pid, NULL, 0);
  return 0;
} /* start_isc */

