#ifndef _LIST_H
#define _LIST_H

#include "defines.h"
#include <sys/types.h>
#include <sys/socket.h>

/* loc_list_t ************************************************************************
 *
 *  - the location_list is a part of each invitation_list entry which stores possible 
 *      locations of invited users
 *  - new location_list entries are added by the "locate_user" procedure or after 
 *     receiving appropriate responses
 *
 *  parts of the structure:
 *  - addr: IP address of the location, identifier of the stucture
 *  - hostname: hostname of the location
 *  - user: user part of the SIP location (user@host)
 *  - retransmissions: counter of transmissions of the request to the location,
 *     used when performing a parallel search or changing of transport (TCP->UDP proxy)
 *  - response_received: flag to indicate if a definitive response was received
 *     from this location; when flag is set (=1) incoming responses from this location are 
 *     ignored
 *  - next: pointer to the next location_list entry
 *  - inv_parent: pointer to the belonging invitation_list entry 
 *
 ***********************************************************************************/
typedef struct loc_list_t {
  char			addr[17]; /* IP Addr */
  char			hostname[MAX_HOST_NAME_LENGTH];
  char			user[MAX_USER_NAME_LENGTH]; /* user login */
  int                     retransmissions; /* counter for request retransmissions */
  int response_received; /* flag*/
  struct loc_list_t		*next; /* pointer to next entry */
  struct inv_list_t         *inv_parent;  /* ponter to invitation_list entry */              
} loc_list_t;



/* inv_list_t ************************************************************************
 *
 *  - the invitation_list stores invitations during their processing
 *  - new invitation_list entries are added when receiving a new SIP request and are
 *    removed after sending a definite response
 *
 *  parts of the structure:
 *  - call_id: call-id of the invitation, identifier of the stucture
 *  - request: incoming SIP request
 *  - user: user specified in the first SIP request line
 *  - userEntry: pointer to invited user's passwd structure
 *
 *  - response: received SIP response with highest priority
 *  - response_priority: priority of the response stored in response
 *
 *  - addr_of_client: IP address of the client from which a request was received
 *  - socket_to_client: socket to the client
 *  - transport_to_client: transport protocol to the client (SOCK_STREAM or 
 *    SOCK_DGRAM)
 *  - socket_to_server: socket to the next SIP server when performing TCP proxy mode
 *  - transport_to_server: transport protocol to the next SIP server (SOCK_STREAM)
 *
 *  - state: current state of the request
 *  - locations: - during user location: number of found locations of the user
 *               - counter of locations with outstanding definite responses
 *  - location_list: pointer to the invitation's location list
 *  - next: pointer to the next invitation_list entry
 *
 ***********************************************************************************/
typedef struct inv_list_t {
  char call_id[MAX_CALL_ID_LENGTH];
  char user[MAX_USER_NAME_LENGTH];
  struct passwd *userEntry;
  char request[MSG_BUF_SIZE];
  char response[MSG_BUF_SIZE];
  int response_priority;
  char addr_of_client[17];
  int  socket_to_client;        
  int  port_to_client;          
  int  transport_to_client;     
  int  socket_to_server;        
  int  transport_to_server;     
  int state;                   /* current status of this invitation */
  int locations;
  loc_list_t *location_list;
  struct inv_list_t *next;

 } inv_list_t;


extern int loc_list_add (inv_list_t * inv);
extern int inv_list_add ();
extern void inv_list_remove (char *str);
extern inv_list_t *inv_list_search (char *search_str);
extern loc_list_t *loc_list_search (inv_list_t *inv, char *search_str);

#endif /* _LIST_H */
