/***********************************************************************************
 *
 * functions to perform TCP communication
 *
 * Copyright 1997 by Stefan Hoffmann
 * GMD Fokus
 *
 ***********************************************************************************/



#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>
#include "debug.h"
#include "list.h"
#include "misc.h"
#include "reply_codes.h"
#include "response.h"



extern int            sip_port;
extern inv_list_t     *invitation_list;
int                   timeout;



void alarm_handler ()
{
  DEBUG (3, syslog (LOG_DEBUG, "TCP_read: Timeout");)
  timeout = TRUE;
} /* alarm_handler */



/* TCP_read ***********************************************************************
 *
 *  read new incoming TCP message
 *  - reads until receiving an empty line which terminates the SIP header
 *  - tries to extract payload length (Content-Length:)
 *  - reads SDP payload
 *
 *  in:
 *  - fd: socket to read from
 *  - inv: pointer to entry in invitation_list or NULL if new invitation
 *
 *  returns:
 *  - pointer to new message
 *  - NULL, if error
 *
 ***********************************************************************************/
char *TCP_read (int fd, inv_list_t *inv)
{
  static char     func[] = "TCP_read";
  FILE            *f;
  char            line[MAX_LINE_LENGTH];
  char            *msg; /* pointer to incoming mesage */
  char            call_id[MAX_CALL_ID_LENGTH];
  char            *payload_len; /* length of the SDP payload within the request */
  char            *p;
  int             i;


  DEBUG (2, syslog (LOG_DEBUG, "%s %d",func, fd);)

  /* get some memory for holding the message */
  if ((msg = malloc (MSG_BUF_SIZE+1)) == NULL) {
    if (inv != NULL)
      reply(inv, SERVER_INTERNAL_ERROR, "Server Internal Error");
      return NULL;
  }
  *msg='\0'; /* malloc do not fill the allocated memory this is to avoid
		garbage in strcat */
  
  /* install timer to avoid blocking */
  timeout = FALSE;
  signal (SIGALRM, alarm_handler);
  alarm(10);

  if ((f = fdopen (fd, "rw")) == NULL) {
    if (inv != NULL)
      reply(inv, SERVER_INTERNAL_ERROR, "Server Internal Error"); 
    free(msg);
    return NULL;
  }
  
  /***********************************************************************************
   * read msg
   ***********************************************************************************/
  while (!timeout) {
    for (i = MSG_BUF_SIZE; i > 0; i -= strlen(line)) {
      if (fgets (line, MAX_LINE_LENGTH, f)) {
        strcat(msg, line);
        if(!strcmp(line, "\r\n")) {
          /***********************************************************************************
           * received an empty line, SIP header is complete
           * get payload length 
           ***********************************************************************************/
        if ((payload_len = search_SIP_header(msg, "\nContent-length:", "\nl:")) == NULL) {
          /* no length field found, set len = 0 */
          i = 0;
        }
        else 
          i = atoi(payload_len);
          if ((strlen(msg) + i) > MSG_BUF_SIZE) {
            /* BAD REQ, too long */
            if (inv != NULL && fd == inv->socket_to_client) 
              reply(inv, BAD_REQUEST, "Bad Request");
            if (inv == NULL) {
              /* try to get call_id to send a reply */
              if ((p = search_SIP_header(msg, "\nCall-id:","\ni:")) != NULL) {
                if ((p = get_field (p, call_id, sizeof(call_id)-1)) != NULL) {
                  if (!inv_list_add()) {
                    strcpy(invitation_list->call_id, call_id);
                    strcpy(invitation_list->request, msg);
                    invitation_list->transport_to_client=SOCK_STREAM;
                    invitation_list->socket_to_client=fd;
                    reply(inv, BAD_REQUEST, "Bad Request");
                  }
                }
              }
	    } 
            free(msg);
            fclose(f);
            return NULL;
          }
        }
      }
      else {
        /* connection closed */
        DEBUG (2, syslog (LOG_DEBUG, "CONNECTION CLOSED");)
        i = -1;
        alarm(0);
        timeout = FALSE;
        fclose(f);
        free(msg);
        if (inv != NULL) {
          /* start timer to delete entry */
          if (fd == inv->socket_to_server) {
            /* if socket to next server was closed before receiving a response send "Search Failure" to client */
            if (inv->state < REPLY_SENT)
              reply(inv, SEARCH_FAILURE, "Search Failure");
          }
          close_event_handler(inv);
        }
        return NULL;
      }
    } /* for */
    break;
  } /* while */
  
  if (timeout) {
    free(msg);
    return NULL;
  }
  alarm(0);
    
  DEBUG (3, syslog (LOG_DEBUG, "%s msg: %s",func, msg);)

  return (msg);
} /* TCP_read */



/* TCP_send ************************************************************************
 *
 *  send a TCP message
 *  in:
 *  -sock: socket to send to
 *  -msg: msessage to send
 *  
 *  returns:
 *  - number of bytes sent
 *
 ***********************************************************************************/
int TCP_send (int sock, char *msg)
{
  static char  func[] = "TCP_send";

  
  DEBUG (2, syslog (LOG_DEBUG, "%s", func);) 
  DEBUG (3, syslog (LOG_DEBUG, "%s msg: %s", func, msg);) 

  return(send(sock, msg, strlen(msg), 0));
} /* TCP_send */



/* TCP_connect *********************************************************************
 *
 *  establish TCP connection to next sip server
 *
 *  in:  
 *  - loc: pointer to entry in locations_list
 *
 *  retruns:
 *  - socket to next server
 *  - -1 on failure
 *
 ***********************************************************************************/
int TCP_connect (loc_list_t *loc)
{
  static char          func[] = "TCP_connect";
  int                  size = SOCK_BUF_SIZE;
  int                  sock; /* new socket to next server */
  unsigned long        addr; /* address of next server */
  struct sockaddr_in   sin;
 

  DEBUG (2, syslog (LOG_DEBUG, "%s",func);)

  if ((sock = socket(AF_INET, SOCK_STREAM, 0)) == -1) {
    DEBUG (1, syslog (LOG_ERR, "%s: sock: %m", func);)
    return -1;
  }
  
  if (setsockopt (sock, SOL_SOCKET,SO_SNDBUF,(char *)&size, sizeof(int)) < 0) {
    DEBUG (1, syslog (LOG_ERR, "%s: setsockopt: %m", func);) 
    close (sock);
    return(-1);
  }
  
  if (setsockopt (sock, SOL_SOCKET,SO_RCVBUF,(char *)&size, sizeof(int)) < 0) {
    DEBUG (1, syslog (LOG_ERR, "%s: setsockopt: %m", func);)
    close (sock);
    return(-1);
  }
  
  addr = inet_addr(loc->addr);
 
  sin.sin_addr.s_addr = htonl(addr);
  sin.sin_port = htons(sip_port);
  sin.sin_family=AF_INET;

  if (connect(sock, (struct sockaddr *)&sin, sizeof(sin)) < 0) {
    DEBUG (1, syslog (LOG_ERR, "%s: connect: %m", func);)
    close (sock);
    return -1;
  }
  
  return sock;
} /* TCP_connect */
